"""
Execute puppet routines
"""

import datetime
import logging
import os

import salt.utils.args
import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
import salt.utils.yaml
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if puppet is installed
    """
    unavailable_exes = ", ".join(
        exe for exe in ("facter", "puppet") if salt.utils.path.which(exe) is None
    )
    if unavailable_exes:
        return (
            False,
            "The puppet execution module cannot be loaded: {} unavailable.".format(
                unavailable_exes
            ),
        )
    else:
        return "puppet"


def _format_fact(output):
    try:
        fact, value = output.split(" => ", 1)
        value = value.strip()
    except ValueError:
        fact = None
        value = None
    return (fact, value)


class _Puppet:
    """
    Puppet helper class. Used to format command for execution.
    """

    def __init__(self):
        """
        Setup a puppet instance, based on the premis that default usage is to
        run 'puppet agent --test'. Configuration and run states are stored in
        the default locations.
        """
        self.subcmd = "agent"
        self.subcmd_args = []  # e.g. /a/b/manifest.pp

        self.kwargs = {"color": "false"}  # e.g. --tags=apache::server
        self.args = []  # e.g. --noop

        puppet_config = __salt__["cmd.run"](
            "puppet config print --render-as yaml vardir rundir confdir"
        )
        conf = salt.utils.yaml.safe_load(puppet_config)
        self.vardir = conf["vardir"]
        self.rundir = conf["rundir"]
        self.confdir = conf["confdir"]

        self.disabled_lockfile = self.vardir + "/state/agent_disabled.lock"
        self.run_lockfile = self.vardir + "/state/agent_catalog_run.lock"
        self.agent_pidfile = self.rundir + "/agent.pid"
        self.lastrunfile = self.vardir + "/state/last_run_summary.yaml"

    def __repr__(self):
        """
        Format the command string to executed using cmd.run_all.
        """
        cmd = "puppet {subcmd} --vardir {vardir} --confdir {confdir}".format(
            **self.__dict__
        )

        args = " ".join(self.subcmd_args)
        args += "".join([f" --{k}" for k in self.args])  # single spaces
        args += "".join([f" --{k} {v}" for k, v in self.kwargs.items()])

        # Ensure that the puppet call will return 0 in case of exit code 2
        if salt.utils.platform.is_windows():
            return "cmd /V:ON /c {} {} ^& if !ERRORLEVEL! EQU 2 (EXIT 0) ELSE (EXIT /B)".format(
                cmd, args
            )
        return f"({cmd} {args}) || test $? -eq 2"

    def arguments(self, args=None):
        """
        Read in arguments for the current subcommand. These are added to the
        cmd line without '--' appended. Any others are redirected as standard
        options with the double hyphen prefixed.
        """
        # permits deleting elements rather than using slices
        args = args and list(args) or []

        # match against all known/supported subcmds
        if self.subcmd == "apply":
            # apply subcommand requires a manifest file to execute
            self.subcmd_args = [args[0]]
            del args[0]

        if self.subcmd == "agent":
            # no arguments are required
            args.extend(["test"])

        # finally do this after subcmd has been matched for all remaining args
        self.args = args


def run(*args, **kwargs):
    """
    Execute a puppet run and return a dict with the stderr, stdout,
    return code, etc. The first positional argument given is checked as a
    subcommand. Following positional arguments should be ordered with arguments
    required by the subcommand first, followed by non-keyword arguments.
    Tags are specified by a tag keyword and comma separated list of values. --
    http://docs.puppetlabs.com/puppet/latest/reference/lang_tags.html

    CLI Examples:

    .. code-block:: bash

        salt '*' puppet.run
        salt '*' puppet.run tags=basefiles::edit,apache::server
        salt '*' puppet.run agent onetime no-daemonize no-usecacheonfailure no-splay ignorecache
        salt '*' puppet.run debug
        salt '*' puppet.run apply /a/b/manifest.pp modulepath=/a/b/modules tags=basefiles::edit,apache::server
    """
    puppet = _Puppet()

    # new args tuple to filter out agent/apply for _Puppet.arguments()
    buildargs = ()
    for arg in args:
        # based on puppet documentation action must come first. making the same
        # assertion. need to ensure the list of supported cmds here matches
        # those defined in _Puppet.arguments()
        if arg in ["agent", "apply"]:
            puppet.subcmd = arg
        else:
            buildargs += (arg,)
    # args will exist as an empty list even if none have been provided
    puppet.arguments(buildargs)

    puppet.kwargs.update(salt.utils.args.clean_kwargs(**kwargs))

    ret = __salt__["cmd.run_all"](repr(puppet), python_shell=True)
    return ret


def noop(*args, **kwargs):
    """
    Execute a puppet noop run and return a dict with the stderr, stdout,
    return code, etc. Usage is the same as for puppet.run.

    CLI Example:

    .. code-block:: bash

        salt '*' puppet.noop
        salt '*' puppet.noop tags=basefiles::edit,apache::server
        salt '*' puppet.noop debug
        salt '*' puppet.noop apply /a/b/manifest.pp modulepath=/a/b/modules tags=basefiles::edit,apache::server
    """
    args += ("noop",)
    return run(*args, **kwargs)


def enable():
    """
    .. versionadded:: 2014.7.0

    Enable the puppet agent

    CLI Example:

    .. code-block:: bash

        salt '*' puppet.enable
    """
    puppet = _Puppet()

    if os.path.isfile(puppet.disabled_lockfile):
        try:
            os.remove(puppet.disabled_lockfile)
        except OSError as exc:
            msg = f"Failed to enable: {exc}"
            log.error(msg)
            raise CommandExecutionError(msg)
        else:
            return True
    return False


def disable(message=None):
    """
    .. versionadded:: 2014.7.0

    Disable the puppet agent

    message
        .. versionadded:: 2015.5.2

        Disable message to send to puppet

    CLI Example:

    .. code-block:: bash

        salt '*' puppet.disable
        salt '*' puppet.disable 'Disabled, contact XYZ before enabling'
    """

    puppet = _Puppet()

    if os.path.isfile(puppet.disabled_lockfile):
        return False
    else:
        with salt.utils.files.fopen(puppet.disabled_lockfile, "w") as lockfile:
            try:
                # Puppet chokes when no valid json is found
                msg = (
                    f'{{"disabled_message":"{message}"}}'
                    if message is not None
                    else "{}"
                )
                lockfile.write(salt.utils.stringutils.to_str(msg))
                lockfile.close()
                return True
            except OSError as exc:
                msg = f"Failed to disable: {exc}"
                log.error(msg)
                raise CommandExecutionError(msg)


def status():
    """
    .. versionadded:: 2014.7.0

    Display puppet agent status

    CLI Example:

    .. code-block:: bash

        salt '*' puppet.status
    """
    puppet = _Puppet()

    if os.path.isfile(puppet.disabled_lockfile):
        return "Administratively disabled"

    if os.path.isfile(puppet.run_lockfile):
        try:
            with salt.utils.files.fopen(puppet.run_lockfile, "r") as fp_:
                pid = int(salt.utils.stringutils.to_unicode(fp_.read()))
                os.kill(pid, 0)  # raise an OSError if process doesn't exist
        except (OSError, ValueError):
            return "Stale lockfile"
        else:
            return "Applying a catalog"

    if os.path.isfile(puppet.agent_pidfile):
        try:
            with salt.utils.files.fopen(puppet.agent_pidfile, "r") as fp_:
                pid = int(salt.utils.stringutils.to_unicode(fp_.read()))
                os.kill(pid, 0)  # raise an OSError if process doesn't exist
        except (OSError, ValueError):
            return "Stale pidfile"
        else:
            return "Idle daemon"

    return "Stopped"


def summary():
    """
    .. versionadded:: 2014.7.0

    Show a summary of the last puppet agent run

    CLI Example:

    .. code-block:: bash

        salt '*' puppet.summary
    """

    puppet = _Puppet()

    try:
        with salt.utils.files.fopen(puppet.lastrunfile, "r") as fp_:
            report = salt.utils.yaml.safe_load(fp_)
        result = {}

        if "time" in report:
            try:
                result["last_run"] = datetime.datetime.fromtimestamp(
                    int(report["time"]["last_run"])
                ).isoformat()
            except (TypeError, ValueError, KeyError):
                result["last_run"] = "invalid or missing timestamp"

            result["time"] = {}
            for key in ("total", "config_retrieval"):
                if key in report["time"]:
                    result["time"][key] = report["time"][key]

        if "resources" in report:
            result["resources"] = report["resources"]

    except salt.utils.yaml.YAMLError as exc:
        raise CommandExecutionError(f"YAML error parsing puppet run summary: {exc}")
    except OSError as exc:
        raise CommandExecutionError(f"Unable to read puppet run summary: {exc}")

    return result


def plugin_sync():
    """
    Runs a plugin sync between the puppet master and agent

    CLI Example:

    .. code-block:: bash

        salt '*' puppet.plugin_sync
    """
    ret = __salt__["cmd.run"]("puppet plugin download")

    if not ret:
        return ""
    return ret


def facts(puppet=False):
    """
    Run facter and return the results

    CLI Example:

    .. code-block:: bash

        salt '*' puppet.facts
    """
    ret = {}
    opt_puppet = "--puppet" if puppet else ""
    cmd_ret = __salt__["cmd.run_all"](f"facter {opt_puppet}")

    if cmd_ret["retcode"] != 0:
        raise CommandExecutionError(cmd_ret["stderr"])

    output = cmd_ret["stdout"]

    # Loop over the facter output and  properly
    # parse it into a nice dictionary for using
    # elsewhere
    for line in output.splitlines():
        if not line:
            continue
        fact, value = _format_fact(line)
        if not fact:
            continue
        ret[fact] = value
    return ret


def fact(name, puppet=False):
    """
    Run facter for a specific fact

    CLI Example:

    .. code-block:: bash

        salt '*' puppet.fact kernel
    """
    opt_puppet = "--puppet" if puppet else ""
    ret = __salt__["cmd.run_all"](f"facter {opt_puppet} {name}", python_shell=False)

    if ret["retcode"] != 0:
        raise CommandExecutionError(ret["stderr"])

    if not ret["stdout"]:
        return ""
    return ret["stdout"]



"""
Connection library for Amazon S3

:depends: requests
"""

import logging
import urllib.parse
import xml.etree.ElementTree as ET

import salt.utils.aws
import salt.utils.files
import salt.utils.hashutils
import salt.utils.xmlutil as xml
from salt.exceptions import CommandExecutionError

try:
    import requests

    HAS_REQUESTS = True  # pylint: disable=W0612
except ImportError:
    HAS_REQUESTS = False  # pylint: disable=W0612


log = logging.getLogger(__name__)


def query(
    key,
    keyid,
    method="GET",
    params=None,
    headers=None,
    requesturl=None,
    return_url=False,
    bucket=None,
    service_url=None,
    path="",
    return_bin=False,
    action=None,
    local_file=None,
    verify_ssl=True,
    full_headers=False,
    kms_keyid=None,
    location=None,
    role_arn=None,
    chunk_size=16384,
    path_style=False,
    https_enable=True,
):
    """
    Perform a query against an S3-like API. This function requires that a
    secret key and the id for that key are passed in. For instance:

        s3.keyid: GKTADJGHEIQSXMKKRBJ08H
        s3.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    If keyid or key is not specified, an attempt to fetch them from EC2 IAM
    metadata service will be made.

    A service_url may also be specified in the configuration:

        s3.service_url: s3.amazonaws.com

    If a service_url is not specified, the default is s3.amazonaws.com. This
    may appear in various documentation as an "endpoint". A comprehensive list
    for Amazon S3 may be found at::

        http://docs.aws.amazon.com/general/latest/gr/rande.html#s3_region

    The service_url will form the basis for the final endpoint that is used to
    query the service.

    Path style can be enabled:

        s3.path_style: True

    This can be useful if you need to use salt with a proxy for an s3 compatible storage

    You can use either https protocol or http protocol:

        s3.https_enable: True

    SSL verification may also be turned off in the configuration:

        s3.verify_ssl: False

    This is required if using S3 bucket names that contain a period, as
    these will not match Amazon's S3 wildcard certificates. Certificate
    verification is enabled by default.

    A region may be specified:

        s3.location: eu-central-1

    If region is not specified, an attempt to fetch the region from EC2 IAM
    metadata service will be made. Failing that, default is us-east-1
    """
    if not HAS_REQUESTS:
        log.error("There was an error: requests is required for s3 access")

    if not headers:
        headers = {}

    if not params:
        params = {}

    if not service_url:
        service_url = "s3.amazonaws.com"

    if not bucket or path_style:
        endpoint = service_url
    else:
        endpoint = f"{bucket}.{service_url}"

    if path_style and bucket:
        path = f"{bucket}/{path}"

    # Try grabbing the credentials from the EC2 instance IAM metadata if available
    if not key:
        key = salt.utils.aws.IROLE_CODE

    if not keyid:
        keyid = salt.utils.aws.IROLE_CODE

    if kms_keyid is not None and method in ("PUT", "POST"):
        headers["x-amz-server-side-encryption"] = "aws:kms"
        headers["x-amz-server-side-encryption-aws-kms-key-id"] = kms_keyid

    if not location:
        location = salt.utils.aws.get_location()

    data = ""
    fh = None
    payload_hash = None
    if method == "PUT":
        if local_file:
            payload_hash = salt.utils.hashutils.get_hash(local_file, form="sha256")

    if path is None:
        path = ""
    path = urllib.parse.quote(path)

    if not requesturl:
        requesturl = "{}://{}/{}".format(
            "https" if https_enable else "http", endpoint, path
        )
        headers, requesturl = salt.utils.aws.sig4(
            method,
            endpoint,
            params,
            data=data,
            uri=f"/{path}",
            prov_dict={"id": keyid, "key": key},
            role_arn=role_arn,
            location=location,
            product="s3",
            requesturl=requesturl,
            headers=headers,
            payload_hash=payload_hash,
        )

    log.debug("S3 Request: %s", requesturl)
    log.debug("S3 Headers::")
    log.debug("    Authorization: %s", headers["Authorization"])

    if not data:
        data = None

    try:
        if method == "PUT":
            if local_file:
                # pylint: disable=resource-leakage
                fh = salt.utils.files.fopen(local_file, "rb")
                # pylint: enable=resource-leakage
                data = fh.read()  # pylint: disable=resource-leakage
            result = requests.request(
                method,
                requesturl,
                headers=headers,
                data=data,
                verify=verify_ssl,
                stream=True,
                timeout=300,
            )
        elif method == "GET" and local_file and not return_bin:
            result = requests.request(
                method,
                requesturl,
                headers=headers,
                data=data,
                verify=verify_ssl,
                stream=True,
                timeout=300,
            )
        else:
            result = requests.request(
                method,
                requesturl,
                headers=headers,
                data=data,
                verify=verify_ssl,
                timeout=300,
            )
    finally:
        if fh is not None:
            fh.close()

    err_code = None
    err_msg = None
    if result.status_code >= 400:
        # On error the S3 API response should contain error message
        err_text = result.content or "Unknown error"
        log.debug("    Response content: %s", err_text)

        # Try to get err info from response xml
        try:
            err_data = xml.to_dict(ET.fromstring(err_text))
            err_code = err_data["Code"]
            err_msg = err_data["Message"]
        except (KeyError, ET.ParseError) as err:
            log.debug(
                "Failed to parse s3 err response. %s: %s", type(err).__name__, err
            )
            err_code = f"http-{result.status_code}"
            err_msg = err_text

    log.debug("S3 Response Status Code: %s", result.status_code)

    if method == "PUT":
        if result.status_code != 200:
            if local_file:
                raise CommandExecutionError(
                    "Failed to upload from {} to {}. {}: {}".format(
                        local_file, path, err_code, err_msg
                    )
                )
            raise CommandExecutionError(
                f"Failed to create bucket {bucket}. {err_code}: {err_msg}"
            )

        if local_file:
            log.debug("Uploaded from %s to %s", local_file, path)
        else:
            log.debug("Created bucket %s", bucket)
        return

    if method == "DELETE":
        if not str(result.status_code).startswith("2"):
            if path:
                raise CommandExecutionError(
                    "Failed to delete {} from bucket {}. {}: {}".format(
                        path, bucket, err_code, err_msg
                    )
                )
            raise CommandExecutionError(
                f"Failed to delete bucket {bucket}. {err_code}: {err_msg}"
            )

        if path:
            log.debug("Deleted %s from bucket %s", path, bucket)
        else:
            log.debug("Deleted bucket %s", bucket)
        return

    # This can be used to save a binary object to disk
    if local_file and method == "GET":
        if result.status_code < 200 or result.status_code >= 300:
            raise CommandExecutionError(f"Failed to get file. {err_code}: {err_msg}")

        log.debug("Saving to local file: %s", local_file)
        with salt.utils.files.fopen(local_file, "wb") as out:
            for chunk in result.iter_content(chunk_size=chunk_size):
                out.write(chunk)
        return f"Saved to local file: {local_file}"

    if result.status_code < 200 or result.status_code >= 300:
        raise CommandExecutionError(f"Failed s3 operation. {err_code}: {err_msg}")

    # This can be used to return a binary object wholesale
    if return_bin:
        return result.content

    if result.content:
        items = ET.fromstring(result.content)

        ret = []
        for item in items:
            ret.append(xml.to_dict(item))

        if return_url is True:
            return ret, requesturl
    else:
        if result.status_code != requests.codes.ok:
            return
        ret = {"headers": []}
        if full_headers:
            ret["headers"] = dict(result.headers)
        else:
            for header in result.headers:
                ret["headers"].append(header.strip())

    return ret



"""
Manage RDP Service on Windows servers
"""


def __virtual__():
    """
    Load only if network_win is loaded
    """
    if "rdp.enable" in __salt__:
        return "rdp"
    return (False, "rdp module could not be loaded")


def enabled(name):
    """
    Enable the RDP service and make sure access to the RDP
    port is allowed in the firewall configuration
    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    stat = __salt__["rdp.status"]()

    if not stat:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "RDP will be enabled"
            return ret

        ret["result"] = __salt__["rdp.enable"]()
        ret["changes"] = {"RDP was enabled": True}
        return ret

    ret["comment"] = "RDP is enabled"
    return ret


def disabled(name):
    """
    Disable the RDP service
    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    stat = __salt__["rdp.status"]()

    if stat:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "RDP will be disabled"
            return ret

        ret["result"] = __salt__["rdp.disable"]()
        ret["changes"] = {"RDP was disabled": True}
        return ret

    ret["comment"] = "RDP is disabled"
    return ret



"""
Integration tests for the lxd states
"""

import pytest

import salt.modules.lxd
from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin


@pytest.mark.flaky(max_runs=4)
@pytest.mark.skipif(salt.modules.lxd.HAS_PYLXD is False, reason="pylxd not installed")
@pytest.mark.skip_if_binaries_missing("lxd", reason="LXD not installed")
@pytest.mark.skip_if_binaries_missing("lxc", reason="LXC not installed")
@pytest.mark.slow_test
class LxdContainerTestCase(ModuleCase, SaltReturnAssertsMixin):
    def setUp(self):
        self.run_state(
            "lxd_image.present",
            name="images:centos/7",
            source={
                "name": "centos/7",
                "type": "simplestreams",
                "server": "https://images.linuxcontainers.org",
            },
        )

    def tearDown(self):
        self.run_state(
            "lxd_image.absent",
            name="images:centos/7",
        )
        self.run_state("lxd_container.absent", name="test-container", stop=True)

    def test_02__create_container(self):
        ret = self.run_state(
            "lxd_container.present",
            name="test-container",
            running=True,
            source={"type": "image", "alias": "images:centos/7"},
            devices={
                "data1": {"type": "disk", "source": "/tmp", "path": "/mnt/data"},
                "port9000": {
                    "type": "proxy",
                    "listen": "tcp:127.0.0.1:9000",
                    "connect": "tcp:127.0.0.1:9000",
                },
            },
        )
        name = "lxd_container_|-test-container_|-test-container_|-present"
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert (
            ret[name]["changes"]["started"] == 'Started the container "test-container"'
        )

    def test_03__change_container(self):
        self.run_state(
            "lxd_container.present",
            name="test-container",
            running=True,
            source={"type": "image", "alias": "images:centos/7"},
            devices={
                "data1": {"type": "disk", "source": "/tmp", "path": "/mnt/data"},
                "data2": {"type": "disk", "source": "/tmp", "path": "/mnt/data2"},
                "port9000": {
                    "type": "proxy",
                    "listen": "tcp:127.0.0.1:9000",
                    "connect": "tcp:127.0.0.1:9000",
                },
                "port9001": {
                    "type": "proxy",
                    "listen": "tcp:127.0.0.1:9001",
                    "connect": "tcp:127.0.0.1:9001",
                },
                "port9002": {
                    "type": "proxy",
                    "listen": "tcp:127.0.0.1:9002",
                    "connect": "tcp:127.0.0.1:9002",
                },
            },
        )
        ret = self.run_state(
            "lxd_container.present",
            name="test-container",
            running=True,
            source={"type": "image", "alias": "images:centos/7"},
            devices={
                "data1": {"type": "disk", "source": "/tmp", "path": "/mnt/data"},
                "data2": {"type": "disk", "source": "/tmp", "path": "/mnt/data3"},
                "port9000": {
                    "type": "proxy",
                    "listen": "tcp:127.0.0.1:9000",
                    "connect": "tcp:127.0.0.1:9000",
                },
                "port9001": {
                    "type": "proxy",
                    "listen": "tcp:127.0.0.1:9001",
                    "connect": "tcp:127.0.0.1:9009",
                },
                "port9003": {
                    "type": "proxy",
                    "listen": "tcp:127.0.0.1:9003",
                    "connect": "tcp:127.0.0.1:9003",
                },
            },
            restart_on_change=True,
            config=[
                {"key": "boot.autostart", "value": 1},
                {"key": "security.privileged", "value": "1"},
            ],
        )
        name = "lxd_container_|-test-container_|-test-container_|-present"
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert ret[name]["changes"]["config"] == {
            "boot.autostart": 'Added config key "boot.autostart" = "1"',
            "security.privileged": 'Added config key "security.privileged" = "1"',
        }
        assert ret[name]["changes"]["devices"] == {
            "data2": 'Changed device "data2"',
            "port9001": 'Changed device "port9001"',
            "port9002": 'Removed device "port9002"',
            "port9003": 'Added device "port9003"',
        }

    def test_08__running_container(self):
        self.run_state(
            "lxd_container.present",
            name="test-container",
            running=True,
            source={"type": "image", "alias": "images:centos/7"},
        )
        ret = self.run_state(
            "lxd_container.running",
            name="test-container",
        )
        self.assertSaltTrueReturn(ret)
        name = "lxd_container_|-test-container_|-test-container_|-running"
        assert name in ret
        assert not ret[name]["changes"]
        assert (
            ret[name]["comment"] == 'The container "test-container" is already running'
        )
        ret = self.run_state(
            "lxd_container.running",
            name="test-container",
            restart=True,
        )
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert ret[name]["changes"] == {
            "restarted": 'Restarted the container "test-container"'
        }
        assert ret[name]["comment"] == 'Restarted the container "test-container"'

    def test_09__stop_container(self):
        self.run_state(
            "lxd_container.present",
            name="test-container",
            running=True,
            source={"type": "image", "alias": "images:centos/7"},
        )
        ret = self.run_state(
            "lxd_container.stopped",
            name="test-container",
        )
        name = "lxd_container_|-test-container_|-test-container_|-stopped"
        self.assertSaltTrueReturn(ret)
        assert ret[name]["changes"] == {
            "stopped": 'Stopped the container "test-container"'
        }
        ret = self.run_state(
            "lxd_container.stopped",
            name="test-container",
        )
        name = "lxd_container_|-test-container_|-test-container_|-stopped"
        self.assertSaltTrueReturn(ret)
        assert not ret[name]["changes"]

    def test_10__delete_container(self):
        self.run_state(
            "lxd_container.present",
            name="test-container",
            running=True,
            source={"type": "image", "alias": "images:centos/7"},
        )
        ret = self.run_state(
            "lxd_container.absent",
            name="test-container",
        )
        name = "lxd_container_|-test-container_|-test-container_|-absent"
        assert name in ret
        assert ret[name]["result"] is False
        ret = self.run_state(
            "lxd_container.stopped",
            name="test-container",
        )
        name = "lxd_container_|-test-container_|-test-container_|-stopped"
        assert name in ret
        assert ret[name]["result"] is True
        ret = self.run_state(
            "lxd_container.absent",
            name="test-container",
        )
        name = "lxd_container_|-test-container_|-test-container_|-absent"
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert ret[name]["changes"] == {
            "deleted": 'Container "test-container" has been deleted.'
        }



import contextlib
import logging
import re
import string

import pytest

import salt.utils.platform
import salt.utils.pycrypto
from salt.exceptions import SaltInvocationError
from tests.support.mock import patch

passwd = "test_password"
invalid_salt = "thissaltistoolong" * 10
expecteds = {
    "sha512": {
        "hashed": "$6$rounds=65601$goodsalt$lZFhiN5M8RTLd9WKDin50H4lF4F8HGMIdwvKs.nTG7f8F0Y4P447Zb9/E8SkUWjY.K10QT3NuHZNDgc/P/NjT1",
        "salt": "rounds=65601$goodsalt",
        "badsalt": "badsalt",
    },
    "sha256": {
        "hashed": "$5$rounds=53501$goodsalt$W.uoco0wMfGLDOlsbW52E6raFS1Nhj0McfUTj2vORt7",
        "salt": "rounds=53501$goodsalt",
        "badsalt": "badsalt",
    },
    "blowfish": {
        "hashed": "$2b$10$goodsaltgoodsaltgoodsObFfGrJwfV.13QddrZIh2w1ccESmvj8K",
        "salt": "10$goodsaltgoodsaltgoodsa",
        "badsalt": "badsaltbadsaltbadsaltb",
    },
    "md5": {
        "hashed": "$1$goodsalt$4XQMx4a4e1MpBB8xzz.TQ0",
        "salt": "goodsalt",
        "badsalt": "badsalt",
    },
    "crypt": {"hashed": "goVHulDpuGA7w", "salt": "go", "badsalt": "ba"},
}


@pytest.fixture(params=["sha512", "sha256", "blowfish", "md5", "crypt"])
def algorithm(request):
    return request.param


@pytest.mark.skipif(not salt.utils.pycrypto.HAS_CRYPT, reason="crypt not available")
@pytest.mark.parametrize(
    "algorithm, expected",
    [
        ("sha512", expecteds["sha512"]),
        ("sha256", expecteds["sha256"]),
        ("blowfish", expecteds["blowfish"]),
        ("md5", expecteds["md5"]),
        ("crypt", expecteds["crypt"]),
    ],
)
def test_gen_hash_crypt(algorithm, expected):
    """
    Test gen_hash with crypt library
    """
    ret = salt.utils.pycrypto.gen_hash(
        crypt_salt=expected["salt"], password=passwd, algorithm=algorithm
    )
    assert ret == expected["hashed"]

    ret = salt.utils.pycrypto.gen_hash(
        crypt_salt=expected["badsalt"], password=passwd, algorithm=algorithm
    )
    assert ret != expected["hashed"]

    ret = salt.utils.pycrypto.gen_hash(
        crypt_salt=None, password=passwd, algorithm=algorithm
    )
    assert ret != expected["hashed"]


@pytest.mark.skipif(not salt.utils.pycrypto.HAS_CRYPT, reason="crypt not available")
def test_gen_hash_crypt_no_arguments():
    # Assert it works without arguments passed
    assert salt.utils.pycrypto.gen_hash() is not None


@pytest.mark.skipif(not salt.utils.pycrypto.HAS_CRYPT, reason="crypt not available")
def test_gen_hash_crypt_default_algorithm():
    # Assert it works without algorithm passed
    default_algorithm = salt.utils.pycrypto.crypt.methods[0].name.lower()
    expected = expecteds[default_algorithm]
    ret = salt.utils.pycrypto.gen_hash(crypt_salt=expected["salt"], password=passwd)
    assert ret == expected["hashed"]


@pytest.mark.skipif(not salt.utils.pycrypto.HAS_PASSLIB, reason="passlib not available")
@pytest.mark.parametrize(
    "algorithm, expected",
    [
        ("sha512", expecteds["sha512"]),
        ("sha256", expecteds["sha256"]),
        ("blowfish", expecteds["blowfish"]),
        ("md5", expecteds["md5"]),
        ("crypt", expecteds["crypt"]),
    ],
)
def test_gen_hash_passlib(algorithm, expected):
    """
    Test gen_hash with passlib
    """
    with patch("salt.utils.pycrypto.methods", {}):
        with patch("salt.utils.pycrypto.HAS_CRYPT", False):
            ret = salt.utils.pycrypto.gen_hash(
                crypt_salt=expected["salt"], password=passwd, algorithm=algorithm
            )
            assert ret == expected["hashed"]

            ret = salt.utils.pycrypto.gen_hash(
                crypt_salt=expected["badsalt"], password=passwd, algorithm=algorithm
            )
            assert ret != expected["hashed"]

            ret = salt.utils.pycrypto.gen_hash(
                crypt_salt=None, password=passwd, algorithm=algorithm
            )
            assert ret != expected["hashed"]


@pytest.mark.skipif(not salt.utils.pycrypto.HAS_PASSLIB, reason="passlib not available")
def test_gen_hash_passlib_no_arguments():
    # Assert it works without arguments passed
    assert salt.utils.pycrypto.gen_hash() is not None


def test_gen_hash_passlib_default_algorithm():
    # Assert it works without algorithm passed
    default_algorithm = salt.utils.pycrypto.known_methods[0]
    expected = expecteds[default_algorithm]
    if default_algorithm in expected:
        ret = salt.utils.pycrypto.gen_hash(crypt_salt=expected["salt"], password=passwd)
        assert ret == expected["hashed"]


def test_gen_hash_no_lib():
    """
    test gen_hash with no crypt library available
    """
    with patch("salt.utils.pycrypto.HAS_CRYPT", False):
        with patch("salt.utils.pycrypto.HAS_PASSLIB", False):
            with pytest.raises(SaltInvocationError):
                salt.utils.pycrypto.gen_hash()


def test_gen_hash_selection():
    """
    verify the hash backend selection works correctly
    """

    with patch("salt.utils.pycrypto.HAS_CRYPT", True):
        with patch("salt.utils.pycrypto.methods", {"crypt": None}):
            with patch("salt.utils.pycrypto.HAS_PASSLIB", True):
                with patch(
                    "salt.utils.pycrypto._gen_hash_crypt", autospec=True
                ) as gh_crypt:
                    with patch(
                        "salt.utils.pycrypto._gen_hash_passlib", autospec=True
                    ) as gh_passlib:
                        with pytest.raises(SaltInvocationError):
                            salt.utils.pycrypto.gen_hash(algorithm="doesntexist")

                        salt.utils.pycrypto.gen_hash(algorithm="crypt")
                        gh_crypt.assert_called_once()
                        gh_passlib.assert_not_called()

                        gh_crypt.reset_mock()
                        salt.utils.pycrypto.gen_hash(algorithm="sha512")
                        gh_crypt.assert_not_called()
                        gh_passlib.assert_called_once()


def test_gen_hash_crypt_warning(caplog):
    """
    Verify that a bad crypt salt triggers a warning
    """
    with caplog.at_level(logging.WARNING):
        with contextlib.suppress(Exception):
            salt.utils.pycrypto.gen_hash(
                crypt_salt="toolong", password=passwd, algorithm="crypt"
            )
    assert "Hash salt is too long for 'crypt' hash." in caplog.text


def test_secure_password():
    """
    test secure_password
    """
    with patch("salt.utils.pycrypto.HAS_RANDOM", True):
        ret = salt.utils.pycrypto.secure_password()
        check = re.compile(r"[!@#$%^&*()_=+]")
        check_printable = re.compile(
            r"[^{}]".format(
                re.escape(
                    string.ascii_lowercase
                    + string.ascii_uppercase
                    + string.digits
                    + string.punctuation
                )
            )
        )
        check_whitespace = re.compile(rf"[{string.whitespace}]")
        assert check_printable.search(ret) is None
        assert check_whitespace.search(ret) is None
        assert ret
        assert salt.utils.pycrypto.secure_password(length=1, chars="A") == "A"
        assert len(salt.utils.pycrypto.secure_password(length=64)) == 64


def test_secure_password_all_chars():
    """
    test secure_password
    """
    with patch("salt.utils.pycrypto.HAS_RANDOM", True):
        ret = salt.utils.pycrypto.secure_password(
            lowercase=True,
            uppercase=True,
            digits=True,
            punctuation=True,
            whitespace=True,
            printable=True,
        )
        check = re.compile(rf"[^{re.escape(string.printable)}]")
        assert check.search(ret) is None
        assert ret


def test_secure_password_no_has_random():
    """
    test secure_password
    """
    with patch("salt.utils.pycrypto.HAS_RANDOM", False):
        ret = salt.utils.pycrypto.secure_password()
        check_printable = re.compile(
            r"[^{}]".format(
                re.escape(
                    string.ascii_lowercase
                    + string.ascii_uppercase
                    + string.digits
                    + string.punctuation
                )
            )
        )
        check_whitespace = re.compile(rf"[{string.whitespace}]")
        assert check_printable.search(ret) is None
        assert check_whitespace.search(ret) is None
        assert ret
        assert salt.utils.pycrypto.secure_password(length=1, chars="A") == "A"
        assert len(salt.utils.pycrypto.secure_password(length=64)) == 64


def test_secure_password_all_chars_no_has_random():
    """
    test secure_password
    """
    with patch("salt.utils.pycrypto.HAS_RANDOM", False):
        ret = salt.utils.pycrypto.secure_password(printable=True)
        check = re.compile(f"[^{re.escape(string.printable)}]")
        assert check.search(ret) is None
        assert ret



import pytest

import salt.utils.win_dotnet as win_dotnet
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

pytestmark = [
    pytest.mark.skip_unless_on_windows,
]


class WinDotNetTestCase(TestCase):
    """
    Test cases for salt.utils.win_dotnet
    """

    def setUp(self):
        self.mock_reg_list = MagicMock(
            return_value=["CDF", "v2.0.50727", "v3.0", "v3.5", "v4", "v4.0"]
        )
        self.mock_reg_exists = MagicMock(
            side_effect=[
                True,  # Version exists in v2.0.50727
                True,  # Version exists in v3.0
                True,  # Version exists in v3.5
                False,  # Version not in v4
                True,  # Release exists in v4
                False,  # Version not in v4.0
                False,  # Release not in v4.0
            ]
        )
        self.mock_reg_read = MagicMock(
            side_effect=[
                # v2.0.50727
                {"vdata": 1},
                {"vdata": "2.0.50727.4927"},
                {"vdata": 2},
                # v3.0
                {"vdata": 1},
                {"vdata": "3.0.30729.4926"},
                {"vdata": 2},
                # v3.5
                {"vdata": 1},
                {"vdata": "3.5.30729.4926"},
                {"vdata": 1},
                # v4
                {"vdata": 1},
                {"vdata": 461814},
            ]
        )

    def test_versions(self):
        """
        Test the versions function
        """
        expected = {
            "details": {
                "v2.0.50727": {
                    "full": "2.0.50727.4927 SP2",
                    "service_pack": 2,
                    "version": "2.0.50727.4927",
                },
                "v3.0": {
                    "full": "3.0.30729.4926 SP2",
                    "service_pack": 2,
                    "version": "3.0.30729.4926",
                },
                "v3.5": {
                    "full": "3.5.30729.4926 SP1",
                    "service_pack": 1,
                    "version": "3.5.30729.4926",
                },
                "v4": {"full": "4.7.2", "service_pack": "N/A", "version": "4.7.2"},
            },
            "versions": ["2.0.50727.4927", "3.0.30729.4926", "3.5.30729.4926", "4.7.2"],
        }
        with patch("salt.utils.win_reg.list_keys", self.mock_reg_list), patch(
            "salt.utils.win_reg.value_exists", self.mock_reg_exists
        ), patch("salt.utils.win_reg.read_value", self.mock_reg_read):
            result = win_dotnet.versions()
            self.assertDictEqual(result, expected)

    def test_versions_list(self):
        expected = ["2.0.50727.4927", "3.0.30729.4926", "3.5.30729.4926", "4.7.2"]
        with patch("salt.utils.win_reg.list_keys", self.mock_reg_list), patch(
            "salt.utils.win_reg.value_exists", self.mock_reg_exists
        ), patch("salt.utils.win_reg.read_value", self.mock_reg_read):
            result = win_dotnet.versions_list()
            self.assertListEqual(result, expected)

    def test_versions_details(self):
        """
        Test the versions function
        """
        expected = {
            "v2.0.50727": {
                "full": "2.0.50727.4927 SP2",
                "service_pack": 2,
                "version": "2.0.50727.4927",
            },
            "v3.0": {
                "full": "3.0.30729.4926 SP2",
                "service_pack": 2,
                "version": "3.0.30729.4926",
            },
            "v3.5": {
                "full": "3.5.30729.4926 SP1",
                "service_pack": 1,
                "version": "3.5.30729.4926",
            },
            "v4": {"full": "4.7.2", "service_pack": "N/A", "version": "4.7.2"},
        }
        with patch("salt.utils.win_reg.list_keys", self.mock_reg_list), patch(
            "salt.utils.win_reg.value_exists", self.mock_reg_exists
        ), patch("salt.utils.win_reg.read_value", self.mock_reg_read):
            result = win_dotnet.versions_details()
            self.assertDictEqual(result, expected)

    def test_version_atleast_35(self):
        with patch("salt.utils.win_reg.list_keys", self.mock_reg_list), patch(
            "salt.utils.win_reg.value_exists", self.mock_reg_exists
        ), patch("salt.utils.win_reg.read_value", self.mock_reg_read):
            self.assertTrue(win_dotnet.version_at_least("3.5"))

    def test_version_atleast_47(self):
        with patch("salt.utils.win_reg.list_keys", self.mock_reg_list), patch(
            "salt.utils.win_reg.value_exists", self.mock_reg_exists
        ), patch("salt.utils.win_reg.read_value", self.mock_reg_read):
            self.assertTrue(win_dotnet.version_at_least("4.7"))

    def test_version_atleast_472(self):
        with patch("salt.utils.win_reg.list_keys", self.mock_reg_list), patch(
            "salt.utils.win_reg.value_exists", self.mock_reg_exists
        ), patch("salt.utils.win_reg.read_value", self.mock_reg_read):
            self.assertTrue(win_dotnet.version_at_least("4.7.2"))

    def test_version_not_atleast_473(self):
        with patch("salt.utils.win_reg.list_keys", self.mock_reg_list), patch(
            "salt.utils.win_reg.value_exists", self.mock_reg_exists
        ), patch("salt.utils.win_reg.read_value", self.mock_reg_read):
            self.assertFalse(win_dotnet.version_at_least("4.7.3"))



"""
Test for the chocolatey module
"""

import os

import pytest

import salt.modules.chocolatey as chocolatey
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture(scope="module")
def choco_path():
    return "C:\\path\\to\\chocolatey.exe"


@pytest.fixture(scope="module")
def chocolatey_path_pd():
    return os.path.join(
        os.environ.get("ProgramData"), "Chocolatey", "bin", "chocolatey.exe"
    )


@pytest.fixture(scope="module")
def choco_path_pd():
    return os.path.join(os.environ.get("ProgramData"), "Chocolatey", "bin", "choco.exe")


@pytest.fixture(scope="module")
def choco_path_sd():
    return os.path.join(
        os.environ.get("SystemDrive"), "Chocolatey", "bin", "chocolatey.bat"
    )


@pytest.fixture(scope="module")
def mock_false():
    return MagicMock(return_value=False)


@pytest.fixture(scope="module")
def mock_true():
    return MagicMock(return_value=True)


@pytest.fixture()
def configure_loader_modules():
    return {chocolatey: {"__context__": {}, "__salt__": {}}}


def test__clear_context(choco_path):
    """
    Tests _clear_context function
    """
    context = {
        "chocolatey._yes": ["--yes"],
        "chocolatey._path": choco_path,
        "chocolatey._version": "0.9.9",
    }
    with patch.dict(chocolatey.__context__, context):
        chocolatey._clear_context()
        # Did it clear all chocolatey items from __context__?
        assert chocolatey.__context__ == {}


def test__yes_context():
    """
    Tests _yes function when it exists in __context__
    """
    with patch.dict(chocolatey.__context__, {"chocolatey._yes": ["--yes"]}):
        result = chocolatey._yes()
        expected = ["--yes"]
        # Did it return correctly
        assert result == expected
        # Did it populate __context__
        assert chocolatey.__context__["chocolatey._yes"] == expected


def test__yes_version_greater():
    """
    Test _yes when Chocolatey version is greater than 0.9.9
    """
    mock_version = MagicMock(return_value="10.0.0")
    with patch("salt.modules.chocolatey.chocolatey_version", mock_version):
        result = chocolatey._yes()
        expected = ["--yes"]
        # Did it return correctly
        assert result == expected
        # Did it populate __context__
        assert chocolatey.__context__["chocolatey._yes"] == expected


def test__yes_version_less_than():
    """
    Test _yes when Chocolatey version is less than 0.9.9
    """
    mock_version = MagicMock(return_value="0.9.0")
    with patch("salt.modules.chocolatey.chocolatey_version", mock_version):
        result = chocolatey._yes()
        expected = []
        # Did it return correctly
        assert result == expected
        # Did it populate __context__
        assert chocolatey.__context__["chocolatey._yes"] == expected


def test__find_chocolatey_context(choco_path):
    """
    Test _find_chocolatey when it exists in __context__
    """
    with patch.dict(chocolatey.__context__, {"chocolatey._path": choco_path}):
        result = chocolatey._find_chocolatey()
        expected = choco_path
        assert result == expected


def test__find_chocolatey_which(choco_path):
    """
    Test _find_chocolatey when found with `cmd.which`
    """
    mock_which = MagicMock(return_value=choco_path)
    with patch.dict(chocolatey.__salt__, {"cmd.which": mock_which}):
        result = chocolatey._find_chocolatey()
        expected = choco_path
        # Does it return the correct path
        assert result == expected
        # Does it populate __context__
        assert chocolatey.__context__["chocolatey._path"] == expected


def test__find_chocolatey_programdata(mock_false, mock_true, chocolatey_path_pd):
    """
    Test _find_chocolatey when found in ProgramData and named chocolatey.exe
    """
    with patch.dict(chocolatey.__salt__, {"cmd.which": mock_false}), patch(
        "os.path.isfile", mock_true
    ):
        result = chocolatey._find_chocolatey()
        expected = chocolatey_path_pd
        # Does it return the correct path
        assert result == expected
        # Does it populate __context__
        assert chocolatey.__context__["chocolatey._path"] == expected


def test__find_choco_programdata(mock_false, choco_path_pd):
    """
    Test _find_chocolatey when found in ProgramData and named choco.exe
    """
    mock_is_file = MagicMock(side_effect=[False, True])
    with patch.dict(chocolatey.__salt__, {"cmd.which": mock_false}), patch(
        "os.path.isfile", mock_is_file
    ):
        result = chocolatey._find_chocolatey()
        expected = choco_path_pd
        # Does it return the correct path
        assert result == expected
        # Does it populate __context__
        assert chocolatey.__context__["chocolatey._path"] == expected


def test__find_chocolatey_systemdrive(mock_false, choco_path_sd):
    """
    Test _find_chocolatey when found on SystemDrive (older versions)
    """
    with patch.dict(chocolatey.__salt__, {"cmd.which": mock_false}), patch(
        "os.path.isfile", MagicMock(side_effect=[False, False, True])
    ):
        result = chocolatey._find_chocolatey()
        expected = choco_path_sd
        # Does it return the correct path
        assert result == expected
        # Does it populate __context__
        assert chocolatey.__context__["chocolatey._path"] == expected


def test_version_check_remote_false():
    """
    Test version when remote is False
    """
    list_return_value = {"ack": ["3.1.1"]}
    with patch.object(chocolatey, "list_", return_value=list_return_value):
        expected = {"ack": {"installed": ["3.1.1"]}}
        result = chocolatey.version("ack", check_remote=False)
        assert result == expected


def test_version_check_remote_true():
    """
    Test version when remote is True
    """
    list_side_effect = [
        {"ack": ["3.1.1"]},
        {"ack": ["3.1.1"], "Wolfpack": ["3.0.17"], "blackbird": ["1.0.79.3"]},
    ]
    with patch.object(chocolatey, "list_", side_effect=list_side_effect):
        expected = {"ack": {"available": ["3.1.1"], "installed": ["3.1.1"]}}
        result = chocolatey.version("ack", check_remote=True)
        assert result == expected


def test_version_check_remote_true_capitalization():
    """
    Test version when remote is True
    """
    list_side_effect = [
        {"Ack": ["3.1.1"]},
        {"Ack": ["3.1.1"], "Wolfpack": ["3.0.17"], "blackbird": ["1.0.79.3"]},
    ]
    with patch.object(chocolatey, "list_", side_effect=list_side_effect):
        expected = {"ack": {"available": ["3.1.1"], "installed": ["3.1.1"]}}
        result = chocolatey.version("ack", check_remote=True)
        assert result == expected


def test_version_check_remote_true_not_available():
    """
    Test version when remote is True but remote version is unavailable
    """
    list_side_effect = [
        {"ack": ["3.1.1"]},
        {"Wolfpack": ["3.0.17"], "blackbird": ["1.0.79.3"]},
    ]
    with patch.object(chocolatey, "list_", side_effect=list_side_effect):
        expected = {"ack": {"installed": ["3.1.1"]}}
        result = chocolatey.version("ack", check_remote=True)
        assert result == expected


def test_add_source(choco_path):
    """
    Test add_source when remote is False
    """
    cmd_run_all_mock = MagicMock(return_value={"retcode": 0, "stdout": "data"})
    cmd_run_which_mock = MagicMock(return_value=choco_path)
    with patch.dict(
        chocolatey.__salt__,
        {"cmd.which": cmd_run_which_mock, "cmd.run_all": cmd_run_all_mock},
    ):
        expected_call = [
            choco_path,
            "sources",
            "add",
            "--name",
            "source_name",
            "--source",
            "source_location",
        ]

        result = chocolatey.add_source("source_name", "source_location")
        cmd_run_all_mock.assert_called_with(expected_call, python_shell=False)

        expected_call = [
            choco_path,
            "sources",
            "add",
            "--name",
            "source_name",
            "--source",
            "source_location",
            "--priority",
            "priority",
        ]

        result = chocolatey.add_source(
            "source_name", "source_location", priority="priority"
        )
        cmd_run_all_mock.assert_called_with(expected_call, python_shell=False)


def test_list_pre_2_0_0():
    mock_version = MagicMock(return_value="1.2.1")
    mock_find = MagicMock(return_value=choco_path)
    mock_run = MagicMock(return_value={"stdout": "No packages", "retcode": 0})
    with patch.object(chocolatey, "chocolatey_version", mock_version), patch.object(
        chocolatey, "_find_chocolatey", mock_find
    ), patch.dict(chocolatey.__salt__, {"cmd.run_all": mock_run}):
        chocolatey.list_()
        expected_call = [choco_path, "list", "--limit-output"]
        mock_run.assert_called_with(expected_call, python_shell=False)


def test_list_post_2_0_0():
    mock_version = MagicMock(return_value="2.0.1")
    mock_find = MagicMock(return_value=choco_path)
    mock_run = MagicMock(return_value={"stdout": "No packages", "retcode": 0})
    with patch.object(chocolatey, "chocolatey_version", mock_version), patch.object(
        chocolatey, "_find_chocolatey", mock_find
    ), patch.dict(chocolatey.__salt__, {"cmd.run_all": mock_run}):
        chocolatey.list_()
        expected_call = [choco_path, "search", "--limit-output"]
        mock_run.assert_called_with(expected_call, python_shell=False)


def test_list_webpi_pre_2_0_0():
    mock_version = MagicMock(return_value="1.2.1")
    mock_find = MagicMock(return_value=choco_path)
    mock_run = MagicMock(return_value={"stdout": "No packages", "retcode": 0})
    with patch.object(chocolatey, "chocolatey_version", mock_version), patch.object(
        chocolatey, "_find_chocolatey", mock_find
    ), patch.dict(chocolatey.__salt__, {"cmd.run_all": mock_run}):
        chocolatey.list_webpi()
        expected_call = [choco_path, "list", "--source", "webpi"]
        mock_run.assert_called_with(expected_call, python_shell=False)


def test_list_webpi_post_2_0_0():
    mock_version = MagicMock(return_value="2.0.1")
    mock_find = MagicMock(return_value=choco_path)
    mock_run = MagicMock(return_value={"stdout": "No packages", "retcode": 0})
    with patch.object(chocolatey, "chocolatey_version", mock_version), patch.object(
        chocolatey, "_find_chocolatey", mock_find
    ), patch.dict(chocolatey.__salt__, {"cmd.run_all": mock_run}):
        chocolatey.list_webpi()
        expected_call = [choco_path, "search", "--source", "webpi"]
        mock_run.assert_called_with(expected_call, python_shell=False)


def test_list_windowsfeatures_pre_2_0_0():
    mock_version = MagicMock(return_value="1.2.1")
    mock_find = MagicMock(return_value=choco_path)
    mock_run = MagicMock(return_value={"stdout": "No packages", "retcode": 0})
    with patch.object(chocolatey, "chocolatey_version", mock_version), patch.object(
        chocolatey, "_find_chocolatey", mock_find
    ), patch.dict(chocolatey.__salt__, {"cmd.run_all": mock_run}):
        chocolatey.list_windowsfeatures()
        expected_call = [choco_path, "list", "--source", "windowsfeatures"]
        mock_run.assert_called_with(expected_call, python_shell=False)


def test_list_windowsfeatures_post_2_0_0():
    mock_version = MagicMock(return_value="2.0.1")
    mock_find = MagicMock(return_value=choco_path)
    mock_run = MagicMock(return_value={"stdout": "No packages", "retcode": 0})
    with patch.object(chocolatey, "chocolatey_version", mock_version), patch.object(
        chocolatey, "_find_chocolatey", mock_find
    ), patch.dict(chocolatey.__salt__, {"cmd.run_all": mock_run}):
        chocolatey.list_windowsfeatures()
        expected_call = [choco_path, "search", "--source", "windowsfeatures"]
        mock_run.assert_called_with(expected_call, python_shell=False)


def test_chocolatey_version():
    context = {
        "chocolatey._version": "0.9.9",
    }
    with patch.dict(chocolatey.__context__, context):
        result = chocolatey.chocolatey_version()
        expected = "0.9.9"
        assert result == expected


def test_chocolatey_version_refresh():
    context = {"chocolatey._version": "0.9.9"}
    mock_find = MagicMock(return_value="some_path")
    mock_run = MagicMock(return_value="2.2.0")
    with patch.dict(chocolatey.__context__, context), patch.object(
        chocolatey, "_find_chocolatey", mock_find
    ), patch.dict(chocolatey.__salt__, {"cmd.run": mock_run}):
        result = chocolatey.chocolatey_version(refresh=True)
        expected = "2.2.0"
        assert result == expected



"""
Generate baseline proxy minion grains for panos hosts.

"""

import logging

import salt.proxy.panos
import salt.utils.platform

__proxyenabled__ = ["panos"]
__virtualname__ = "panos"

log = logging.getLogger(__file__)

GRAINS_CACHE = {"os_family": "panos"}


def __virtual__():
    try:
        if salt.utils.platform.is_proxy() and __opts__["proxy"]["proxytype"] == "panos":
            return __virtualname__
    except KeyError:
        pass

    return False


def panos(proxy=None):
    if not proxy:
        return {}
    if proxy["panos.initialized"]() is False:
        return {}
    return {"panos": proxy["panos.grains"]()}



import logging
import os
import pathlib
import shutil
import subprocess
import sys

import pytest
import yaml
from pytestskipmarkers.utils import platform
from saltfactories.utils import random_string

import salt.config
import salt.utils.files
from tests.conftest import CODE_DIR, FIPS_TESTRUN
from tests.support.pkg import ApiRequest, SaltMaster, SaltMasterWindows, SaltPkgInstall

log = logging.getLogger(__name__)

# Variable defining a FIPS test run or not
FIPS_TESTRUN = os.environ.get("FIPS_TESTRUN", "0") == "1"


@pytest.fixture(scope="session")
def version(install_salt):
    """
    get version number from artifact
    """
    return install_salt.version


@pytest.fixture(scope="session", autouse=True)
def _system_up_to_date(
    grains,
    shell,
):
    gpg_dest = "/etc/apt/keyrings/salt-archive-keyring.gpg"
    if os.path.exists(gpg_dest):
        with salt.utils.files.fopen(gpg_dest, "r") as fp:
            log.error("Salt gpg key is %s", fp.read())
    else:
        log.error("Salt gpg not present")
    if grains["os_family"] == "Debian":
        ret = shell.run("apt", "update")
        assert ret.returncode == 0
        env = os.environ.copy()
        env["DEBIAN_FRONTEND"] = "noninteractive"
        ret = shell.run(
            "apt",
            "upgrade",
            "-y",
            "-o",
            "DPkg::Options::=--force-confdef",
            "-o",
            "DPkg::Options::=--force-confold",
            env=env,
        )
        assert ret.returncode == 0
    elif grains["os_family"] == "Redhat":
        ret = shell.run("yum", "update", "-y")
        assert ret.returncode == 0


def pytest_addoption(parser):
    """
    register argparse-style options and ini-style config values.
    """
    test_selection_group = parser.getgroup("Tests Runtime Selection")
    test_selection_group.addoption(
        "--pkg-system-service",
        default=False,
        action="store_true",
        help="Run the daemons as system services",
    )
    test_selection_group.addoption(
        "--upgrade",
        default=False,
        action="store_true",
        help="Install previous version and then upgrade then run tests",
    )
    test_selection_group.addoption(
        "--downgrade",
        default=False,
        action="store_true",
        help="Install current version and then downgrade to the previous version and run tests",
    )
    test_selection_group.addoption(
        "--no-install",
        default=False,
        action="store_true",
        help="Do not install salt and use a previous install Salt package",
    )
    test_selection_group.addoption(
        "--no-uninstall",
        default=False,
        action="store_true",
        help="Do not uninstall salt packages after test run is complete",
    )
    test_selection_group.addoption(
        "--prev-version",
        action="store",
        help="Test an upgrade from the version specified.",
    )
    test_selection_group.addoption(
        "--use-prev-version",
        action="store_true",
        help="Tells the test suite to validate the version using the previous version (for downgrades)",
    )
    test_selection_group.addoption(
        "--download-pkgs",
        default=False,
        action="store_true",
        help="Test package download tests",
    )


@pytest.hookimpl(hookwrapper=True, trylast=True)
def pytest_collection_modifyitems(config, items):
    """
    called after collection has been performed, may filter or re-order
    the items in-place.

    :param _pytest.main.Session session: the pytest session object
    :param _pytest.config.Config config: pytest config object
    :param List[_pytest.nodes.Item] items: list of item objects
    """
    # Let PyTest or other plugins handle the initial collection
    yield
    selected = []
    deselected = []
    pkg_tests_path = pathlib.Path(__file__).parent

    if config.getoption("--upgrade"):
        for item in items:
            if str(item.fspath).startswith(str(pkg_tests_path / "upgrade")):
                selected.append(item)
            else:
                deselected.append(item)
    elif config.getoption("--downgrade"):
        for item in items:
            if str(item.fspath).startswith(str(pkg_tests_path / "downgrade")):
                selected.append(item)
            else:
                deselected.append(item)
    elif config.getoption("--download-pkgs"):
        for item in items:
            if str(item.fspath).startswith(str(pkg_tests_path / "download")):
                selected.append(item)
            else:
                deselected.append(item)
    else:
        exclude_paths = (
            str(pkg_tests_path / "upgrade"),
            str(pkg_tests_path / "downgrade"),
            str(pkg_tests_path / "download"),
        )
        for item in items:
            if str(item.fspath).startswith(exclude_paths):
                deselected.append(item)
            else:
                selected.append(item)

    if deselected:
        # Selection changed
        items[:] = selected
        config.hook.pytest_deselected(items=deselected)


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_setup(item):
    """
    Fixtures injection based on markers or test skips based on CLI arguments
    """
    pkg_tests_path = pathlib.Path(__file__).parent
    if (
        str(item.fspath).startswith(str(pkg_tests_path / "download"))
        and item.config.getoption("--download-pkgs") is False
    ):
        raise pytest.skip.Exception(
            "The package download tests are disabled. Pass '--download-pkgs' to pytest "
            "to enable them.",
            _use_item_location=True,
        )

    for key in ("upgrade", "downgrade"):
        if (
            str(item.fspath).startswith(str(pkg_tests_path / key))
            and item.config.getoption(f"--{key}") is False
        ):
            raise pytest.skip.Exception(
                f"The package {key} tests are disabled. Pass '--{key}' to pytest "
                "to enable them.",
                _use_item_location=True,
            )


@pytest.fixture(scope="session")
def salt_factories_root_dir(request, tmp_path_factory):
    root_dir = SaltPkgInstall.salt_factories_root_dir(
        request.config.getoption("--pkg-system-service")
    )
    if root_dir is not None:
        yield root_dir
    else:
        if platform.is_darwin():
            root_dir = pathlib.Path("/tmp/salt-tests-tmpdir")
            root_dir.mkdir(mode=0o777, parents=True, exist_ok=True)
        else:
            root_dir = tmp_path_factory.mktemp("salt-tests")
        try:
            yield root_dir
        finally:
            shutil.rmtree(str(root_dir), ignore_errors=True)


@pytest.fixture(scope="session")
def salt_factories_config(salt_factories_root_dir):
    return {
        "code_dir": CODE_DIR,
        "root_dir": salt_factories_root_dir,
        "system_service": True,
    }


@pytest.fixture(scope="session")
def install_salt(request, salt_factories_root_dir):
    if platform.is_windows():
        conf_dir = "c:/salt/etc/salt"
    else:
        conf_dir = salt_factories_root_dir / "etc" / "salt"
    with SaltPkgInstall(
        conf_dir=conf_dir,
        pkg_system_service=request.config.getoption("--pkg-system-service"),
        upgrade=request.config.getoption("--upgrade"),
        downgrade=request.config.getoption("--downgrade"),
        no_uninstall=request.config.getoption("--no-uninstall"),
        no_install=request.config.getoption("--no-install"),
        prev_version=request.config.getoption("--prev-version"),
        use_prev_version=request.config.getoption("--use-prev-version"),
    ) as fixture:
        yield fixture


@pytest.fixture(scope="session")
def salt_factories(salt_factories, salt_factories_root_dir):
    salt_factories.root_dir = salt_factories_root_dir
    return salt_factories


@pytest.fixture(scope="session")
def salt_master(salt_factories, install_salt, pkg_tests_account):
    """
    Start up a master
    """
    if platform.is_windows():
        state_tree = r"C:\salt\srv\salt"
        pillar_tree = r"C:\salt\srv\pillar"
    elif platform.is_darwin():
        state_tree = "/opt/srv/salt"
        pillar_tree = "/opt/srv/pillar"
    else:
        state_tree = "/srv/salt"
        pillar_tree = "/srv/pillar"

    start_timeout = None
    # Since the daemons are "packaged" with tiamat, the salt plugins provided
    # by salt-factories won't be discovered. Provide the required `*_dirs` on
    # the configuration so that they can still be used.
    config_defaults = {
        "engines_dirs": [
            str(salt_factories.get_salt_engines_path()),
        ],
        "log_handlers_dirs": [
            str(salt_factories.get_salt_log_handlers_path()),
        ],
    }
    if platform.is_darwin():
        config_defaults["enable_fqdns_grains"] = False
    config_overrides = {
        "timeout": 30,
        "file_roots": {
            "base": [
                state_tree,
            ]
        },
        "pillar_roots": {
            "base": [
                pillar_tree,
            ]
        },
        "rest_cherrypy": {
            "port": 8000,
            "disable_ssl": True,
        },
        "netapi_enable_clients": ["local"],
        "external_auth": {
            "auto": {
                pkg_tests_account.username: [
                    ".*",
                ],
            },
        },
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
        "open_mode": True,
    }
    salt_user_in_config_file = False
    master_config = install_salt.config_path / "master"
    if master_config.exists() and master_config.stat().st_size:
        with salt.utils.files.fopen(master_config) as fp:
            data = yaml.safe_load(fp)
            if data is None:
                # File exists but is mostly commented out
                data = {}
            user_in_config_file = data.get("user")
            if user_in_config_file and user_in_config_file == "salt":
                salt_user_in_config_file = True
                # We are testing a different user, so we need to test the system
                # configs, or else permissions will not be correct.
                config_overrides["user"] = user_in_config_file
                config_overrides["log_file"] = salt.config.DEFAULT_MASTER_OPTS.get(
                    "log_file"
                )
                config_overrides["root_dir"] = salt.config.DEFAULT_MASTER_OPTS.get(
                    "root_dir"
                )
                config_overrides["key_logfile"] = salt.config.DEFAULT_MASTER_OPTS.get(
                    "key_logfile"
                )
                config_overrides["pki_dir"] = salt.config.DEFAULT_MASTER_OPTS.get(
                    "pki_dir"
                )
                config_overrides["api_logfile"] = salt.config.DEFAULT_API_OPTS.get(
                    "api_logfile"
                )
                config_overrides["api_pidfile"] = salt.config.DEFAULT_API_OPTS.get(
                    "api_pidfile"
                )
                # verify files were set with correct owner/group
                verify_files = [
                    pathlib.Path("/etc", "salt", "pki", "master"),
                    pathlib.Path("/etc", "salt", "master.d"),
                    pathlib.Path("/var", "cache", "salt", "master"),
                ]
                for _file in verify_files:
                    if _file.owner() != "salt":
                        log.warning(
                            "The owner of '%s' is '%s' when it should be 'salt'",
                            _file,
                            _file.owner(),
                        )
                    if _file.group() != "salt":
                        log.warning(
                            "The group of '%s' is '%s' when it should be 'salt'",
                            _file,
                            _file.group(),
                        )

    master_script = False
    if platform.is_windows():
        master_script = True

    if master_script:
        salt_factories.system_service = False
        salt_factories.generate_scripts = True
        scripts_dir = salt_factories.root_dir / "Scripts"
        scripts_dir.mkdir(exist_ok=True)
        salt_factories.scripts_dir = scripts_dir
        python_executable = install_salt.install_dir / "Scripts" / "python.exe"
        salt_factories.python_executable = python_executable
        factory = salt_factories.salt_master_daemon(
            "pkg-test-master",
            defaults=config_defaults,
            overrides=config_overrides,
            factory_class=SaltMasterWindows,
            salt_pkg_install=install_salt,
        )
        salt_factories.system_service = True
    else:
        factory = salt_factories.salt_master_daemon(
            "pkg-test-master",
            defaults=config_defaults,
            overrides=config_overrides,
            factory_class=SaltMaster,
            salt_pkg_install=install_salt,
        )
    factory.after_terminate(pytest.helpers.remove_stale_master_key, factory)
    if salt_user_in_config_file:
        # Salt factories calls salt.utils.verify.verify_env
        # which sets root perms on /etc/salt/pki/master since we are running
        # the test suite as root, but we want to run Salt master as salt
        # We ensure those permissions where set by the package earlier
        subprocess.run(
            [
                "chown",
                "-R",
                "salt:salt",
                str(pathlib.Path("/etc", "salt", "pki", "master")),
            ],
            check=True,
        )

        if not platform.is_windows() and not platform.is_darwin():
            # The engines_dirs is created in .nox path. We need to set correct perms
            # for the user running the Salt Master
            check_paths = [state_tree, pillar_tree, CODE_DIR / ".nox"]
            for path in check_paths:
                if os.path.exists(path) is False:
                    continue
                subprocess.run(["chown", "-R", "salt:salt", str(path)], check=False)

    with factory.started(start_timeout=start_timeout):
        yield factory


@pytest.fixture(scope="session")
def salt_minion(salt_factories, salt_master, install_salt):
    """
    Start up a minion
    """
    start_timeout = None
    minion_id = random_string("minion-")
    # Since the daemons are "packaged" with tiamat, the salt plugins provided
    # by salt-factories won't be discovered. Provide the required `*_dirs` on
    # the configuration so that they can still be used.
    config_defaults = {
        "engines_dirs": salt_master.config["engines_dirs"].copy(),
        "log_handlers_dirs": salt_master.config["log_handlers_dirs"].copy(),
    }
    if platform.is_darwin():
        config_defaults["enable_fqdns_grains"] = False
    config_overrides = {
        "id": minion_id,
        "file_roots": salt_master.config["file_roots"].copy(),
        "pillar_roots": salt_master.config["pillar_roots"].copy(),
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    if platform.is_windows():
        config_overrides["winrepo_dir"] = (
            rf"{salt_factories.root_dir}\srv\salt\win\repo"
        )
        config_overrides["winrepo_dir_ng"] = (
            rf"{salt_factories.root_dir}\srv\salt\win\repo_ng"
        )
        config_overrides["winrepo_source_dir"] = r"salt://win/repo_ng"

    factory = salt_master.salt_minion_daemon(
        minion_id,
        overrides=config_overrides,
        defaults=config_defaults,
    )

    # Salt factories calls salt.utils.verify.verify_env
    # which sets root perms on /srv/salt and /srv/pillar since we are running
    # the test suite as root, but we want to run Salt master as salt
    if not platform.is_windows() and not platform.is_darwin():
        import pwd

        try:
            pwd.getpwnam("salt")
        except KeyError:
            # The salt user does not exist
            pass
        else:
            state_tree = "/srv/salt"
            pillar_tree = "/srv/pillar"
            check_paths = [state_tree, pillar_tree, CODE_DIR / ".nox"]
            for path in check_paths:
                if os.path.exists(path) is False:
                    continue
                subprocess.run(["chown", "-R", "salt:salt", str(path)], check=False)

    if platform.is_windows():
        minion_pki = pathlib.Path("c:/salt/etc/salt/pki")
        if minion_pki.exists():
            salt.utils.files.rm_rf(minion_pki)

        # Work around missing WMIC until 3008.10 has been released.
        grainsdir = pathlib.Path("c:/salt/etc/grains")
        grainsdir.mkdir(exist_ok=True)
        shutil.copy(r"salt\grains\disks.py", grainsdir)

    factory.after_terminate(
        pytest.helpers.remove_stale_minion_key, salt_master, factory.id
    )
    with factory.started(start_timeout=start_timeout):
        yield factory


@pytest.fixture(scope="module")
def salt_cli(salt_master):
    return salt_master.salt_cli()


@pytest.fixture(scope="module")
def salt_key_cli(salt_master):
    return salt_master.salt_key_cli()


@pytest.fixture(scope="module")
def salt_call_cli(salt_minion):
    return salt_minion.salt_call_cli()


@pytest.fixture(scope="session")
def pkg_tests_account():
    with pytest.helpers.create_account() as account:
        yield account


@pytest.fixture(scope="module")
def extras_pypath():
    extras_dir = "extras-{}.{}".format(*sys.version_info)
    if platform.is_windows():
        return pathlib.Path(
            os.getenv("ProgramFiles"), "Salt Project", "Salt", extras_dir
        )
    elif platform.is_darwin():
        return pathlib.Path("/opt", "salt", extras_dir)
    else:
        return pathlib.Path("/opt", "saltstack", "salt", extras_dir)


@pytest.fixture(scope="module")
def extras_pypath_bin(extras_pypath):
    return extras_pypath / "bin"


@pytest.fixture(scope="module")
def salt_api(salt_master, install_salt, extras_pypath):
    """
    start up and configure salt_api
    """
    shutil.rmtree(str(extras_pypath), ignore_errors=True)
    start_timeout = None
    factory = salt_master.salt_api_daemon()
    with factory.started(start_timeout=start_timeout):
        yield factory


@pytest.fixture(scope="module")
def api_request(pkg_tests_account, salt_api):
    with ApiRequest(
        port=salt_api.config["rest_cherrypy"]["port"], account=pkg_tests_account
    ) as session:
        yield session



import pytest

import salt.client
from salt.exceptions import SaltInvocationError


@pytest.fixture
def local_client():
    return salt.client.get_local_client()


def test_get_local_client(local_client):
    """
    Test that a local client is created
    """
    assert isinstance(local_client, salt.client.LocalClient)


def test_get_local_client_mopts(master_opts):
    master_opts["rest_cherrypy"] = {"port": 8000}
    local_client = salt.client.get_local_client(mopts=master_opts)
    assert isinstance(local_client, salt.client.LocalClient)
    assert local_client.opts == master_opts


@pytest.mark.parametrize(
    "val, expected",
    ((None, 5), (7, 7), ("9", 9), ("eleven", 5), (["13"], 5)),
)
def test_local_client_get_timeout(local_client, val, expected):
    assert local_client._get_timeout(timeout=val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        ("group1", ["L@spongebob,patrick"]),
        ("group2", ["G@os:squidward"]),
        ("group3", ["(", "G@os:plankton", "and", "(", "L@spongebob,patrick", ")", ")"]),
    ),
)
def test_resolve_nodegroup(master_opts, val, expected):
    master_opts["nodegroups"] = {
        "group1": "L@spongebob,patrick",
        "group2": "G@os:squidward",
        "group3": "G@os:plankton and N@group1",
    }
    local_client = salt.client.get_local_client(mopts=master_opts)
    assert local_client._resolve_nodegroup(val) == expected


def test_resolve_nodegroup_error(master_opts):
    master_opts["nodegroups"] = {
        "group1": "L@spongebob,patrick",
        "group2": "G@os:squidward",
        "group3": "G@os:plankton and N@group1",
    }
    local_client = salt.client.get_local_client(mopts=master_opts)
    with pytest.raises(SaltInvocationError):
        local_client._resolve_nodegroup("missing")


def test_prep_pub(local_client):
    result = local_client._prep_pub(
        tgt="*",
        fun="test.ping",
        arg="",
        tgt_type="glob",
        ret="",
        jid="123",
        timeout=7,
    )
    expected = {
        "arg": "",
        "cmd": "publish",
        "fun": "test.ping",
        "jid": "123",
        "key": "",
        "ret": "",
        "tgt": "*",
        "tgt_type": "glob",
        "user": local_client.salt_user,
    }
    assert result == expected


def test_prep_pub_kwargs(local_client):
    result = local_client._prep_pub(
        tgt="*",
        fun="test.ping",
        arg="",
        tgt_type="glob",
        ret="",
        jid="123",
        timeout=7,
        some_kwarg="spongebob",
    )
    expected = {
        "arg": "",
        "cmd": "publish",
        "fun": "test.ping",
        "jid": "123",
        "key": "",
        "ret": "",
        "tgt": "*",
        "tgt_type": "glob",
        "user": local_client.salt_user,
        "kwargs": {
            "some_kwarg": "spongebob",
        },
    }
    assert result == expected


def test_prep_pub_order_masters(master_opts):
    master_opts["order_masters"] = True
    local_client = salt.client.get_local_client(mopts=master_opts)
    result = local_client._prep_pub(
        tgt="*",
        fun="test.ping",
        arg="",
        tgt_type="glob",
        ret="",
        jid="123",
        timeout=7,
    )
    expected = {
        "arg": "",
        "cmd": "publish",
        "fun": "test.ping",
        "jid": "123",
        "key": "",
        "ret": "",
        "tgt": "*",
        "tgt_type": "glob",
        "to": 7,
        "user": local_client.salt_user,
    }
    assert result == expected


def test_prep_pub_nodegroup(master_opts):
    master_opts["nodegroups"] = {
        "group1": "L@spongebob,patrick",
        "group2": "G@os:squidward",
        "group3": "G@os:plankton and N@group1",
    }
    local_client = salt.client.get_local_client(mopts=master_opts)
    result = local_client._prep_pub(
        tgt="group1",
        fun="test.ping",
        arg="",
        tgt_type="nodegroup",
        ret="",
        jid="123",
        timeout=7,
    )
    expected = {
        "arg": "",
        "cmd": "publish",
        "fun": "test.ping",
        "jid": "123",
        "key": "",
        "ret": "",
        "tgt": "L@spongebob,patrick",
        "tgt_type": "compound",
        "user": local_client.salt_user,
    }
    assert result == expected


def test_prep_pub_compound(local_client):
    result = local_client._prep_pub(
        tgt="spongebob,patrick",
        fun="test.ping",
        arg="",
        tgt_type="compound",
        ret="",
        jid="123",
        timeout=7,
    )
    expected = {
        "arg": "",
        "cmd": "publish",
        "fun": "test.ping",
        "jid": "123",
        "key": "",
        "ret": "",
        "tgt": "spongebob,patrick",
        "tgt_type": "compound",
        "user": local_client.salt_user,
    }
    assert result == expected


def test_prep_pub_compound_nodegroup(master_opts):
    master_opts["nodegroups"] = {
        "group1": "L@spongebob,patrick",
        "group2": "G@os:squidward",
        "group3": "G@os:plankton and N@group1",
    }
    local_client = salt.client.get_local_client(mopts=master_opts)
    result = local_client._prep_pub(
        tgt="N@group1",
        fun="test.ping",
        arg="",
        tgt_type="compound",
        ret="",
        jid="123",
        timeout=7,
    )
    expected = {
        "arg": "",
        "cmd": "publish",
        "fun": "test.ping",
        "jid": "123",
        "key": "",
        "ret": "",
        "tgt": "L@spongebob,patrick",
        "tgt_type": "compound",
        "user": local_client.salt_user,
    }
    assert result == expected


def test_prep_pub_ext_job_cache(master_opts):
    master_opts["ext_job_cache"] = "mysql"
    local_client = salt.client.get_local_client(mopts=master_opts)
    result = local_client._prep_pub(
        tgt="spongebob,patrick",
        fun="test.ping",
        arg="",
        tgt_type="glob",
        ret="",
        jid="123",
        timeout=7,
    )
    expected = {
        "arg": "",
        "cmd": "publish",
        "fun": "test.ping",
        "jid": "123",
        "key": "",
        "ret": "mysql",
        "tgt": "spongebob,patrick",
        "tgt_type": "glob",
        "user": local_client.salt_user,
    }
    assert result == expected


def test_prep_pub_ext_job_cache_existing(master_opts):
    master_opts["ext_job_cache"] = "mysql"
    local_client = salt.client.get_local_client(mopts=master_opts)
    result = local_client._prep_pub(
        tgt="spongebob,patrick",
        fun="test.ping",
        arg="",
        tgt_type="glob",
        ret="postgres",
        jid="123",
        timeout=7,
    )
    expected = {
        "arg": "",
        "cmd": "publish",
        "fun": "test.ping",
        "jid": "123",
        "key": "",
        "ret": "postgres,mysql",
        "tgt": "spongebob,patrick",
        "tgt_type": "glob",
        "user": local_client.salt_user,
    }
    assert result == expected



"""
    Tests of salt.states.gem
"""

import pytest

import salt.states.gem as gem
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {gem: {"__opts__": {"test": False}}}


def test_installed():
    gems = {"foo": ["1.0"], "bar": ["2.0"]}
    gem_list = MagicMock(return_value=gems)
    gem_install_succeeds = MagicMock(return_value=True)
    gem_install_fails = MagicMock(return_value=False)

    with patch.dict(gem.__salt__, {"gem.list": gem_list}):
        with patch.dict(gem.__salt__, {"gem.install": gem_install_succeeds}):
            ret = gem.installed("foo")
            assert ret["result"] is True
            ret = gem.installed("quux")
            assert ret["result"] is True
            gem_install_succeeds.assert_called_once_with(
                "quux",
                pre_releases=False,
                ruby=None,
                runas=None,
                version=None,
                proxy=None,
                rdoc=False,
                source=None,
                ri=False,
                gem_bin=None,
            )

        with patch.dict(gem.__salt__, {"gem.install": gem_install_fails}):
            ret = gem.installed("quux")
            assert ret["result"] is False
            gem_install_fails.assert_called_once_with(
                "quux",
                pre_releases=False,
                ruby=None,
                runas=None,
                version=None,
                proxy=None,
                rdoc=False,
                source=None,
                ri=False,
                gem_bin=None,
            )


def test_installed_version():
    gems = {"foo": ["1.0"], "bar": ["2.0"]}
    gem_list = MagicMock(return_value=gems)
    gem_install_succeeds = MagicMock(return_value=True)

    with patch.dict(gem.__salt__, {"gem.list": gem_list}):
        with patch.dict(gem.__salt__, {"gem.install": gem_install_succeeds}):
            ret = gem.installed("foo", version=">= 1.0")
            assert ret["result"] is True
            assert ret["comment"] == "Installed Gem meets version requirements."


def test_removed():
    gems = ["foo", "bar"]
    gem_list = MagicMock(return_value=gems)
    gem_uninstall_succeeds = MagicMock(return_value=True)
    gem_uninstall_fails = MagicMock(return_value=False)
    with patch.dict(gem.__salt__, {"gem.list": gem_list}):
        with patch.dict(gem.__salt__, {"gem.uninstall": gem_uninstall_succeeds}):
            ret = gem.removed("quux")
            assert ret["result"] is True
            ret = gem.removed("foo")
            assert ret["result"] is True
            gem_uninstall_succeeds.assert_called_once_with(
                "foo", None, runas=None, gem_bin=None
            )

        with patch.dict(gem.__salt__, {"gem.uninstall": gem_uninstall_fails}):
            ret = gem.removed("bar")
            assert ret["result"] is False
            gem_uninstall_fails.assert_called_once_with(
                "bar", None, runas=None, gem_bin=None
            )


def test_sources_add():
    gem_sources = ["http://foo", "http://bar"]
    gem_sources_list = MagicMock(return_value=gem_sources)
    gem_sources_add_succeeds = MagicMock(return_value=True)
    gem_sources_add_fails = MagicMock(return_value=False)
    with patch.dict(gem.__salt__, {"gem.sources_list": gem_sources_list}):
        with patch.dict(gem.__salt__, {"gem.sources_add": gem_sources_add_succeeds}):
            ret = gem.sources_add("http://foo")
            assert ret["result"] is True
            ret = gem.sources_add("http://fui")
            assert ret["result"] is True
            gem_sources_add_succeeds.assert_called_once_with(
                source_uri="http://fui", ruby=None, runas=None
            )
        with patch.dict(gem.__salt__, {"gem.sources_add": gem_sources_add_fails}):
            ret = gem.sources_add("http://fui")
            assert ret["result"] is False
            gem_sources_add_fails.assert_called_once_with(
                source_uri="http://fui", ruby=None, runas=None
            )


def test_sources_remove():
    gem_sources = ["http://foo", "http://bar"]
    gem_sources_list = MagicMock(return_value=gem_sources)
    gem_sources_remove_succeeds = MagicMock(return_value=True)
    gem_sources_remove_fails = MagicMock(return_value=False)
    with patch.dict(gem.__salt__, {"gem.sources_list": gem_sources_list}):
        with patch.dict(
            gem.__salt__, {"gem.sources_remove": gem_sources_remove_succeeds}
        ):
            ret = gem.sources_remove("http://fui")
            assert ret["result"] is True
            ret = gem.sources_remove("http://foo")
            assert ret["result"] is True
            gem_sources_remove_succeeds.assert_called_once_with(
                source_uri="http://foo", ruby=None, runas=None
            )
        with patch.dict(gem.__salt__, {"gem.sources_remove": gem_sources_remove_fails}):
            ret = gem.sources_remove("http://bar")
            assert ret["result"] is False
            gem_sources_remove_fails.assert_called_once_with(
                source_uri="http://bar", ruby=None, runas=None
            )



"""
Display output in a table format
=================================

.. versionadded:: 2017.7.0

The ``table`` outputter displays a sequence of rows as table.

Example output:

.. code-block:: text

    edge01.bjm01:
    ----------
        comment:
        ----------
        out:
        ----------
            ______________________________________________________________________________
            | Active | Interface | Last Move |        Mac        | Moves | Static | Vlan |
            ______________________________________________________________________________
            |  True  |  ae1.900  |    0.0    | 40:A6:77:5A:50:01 |   0   | False  | 111  |
            ______________________________________________________________________________
            |  True  |  ae1.111  |    0.0    | 64:16:8D:32:26:58 |   0   | False  | 111  |
            ______________________________________________________________________________
            |  True  |  ae1.111  |    0.0    | 8C:60:4F:73:2D:57 |   0   | False  | 111  |
            ______________________________________________________________________________
            |  True  |  ae1.111  |    0.0    | 8C:60:4F:73:2D:7C |   0   | False  | 111  |
            ______________________________________________________________________________
            |  True  |  ae1.222  |    0.0    | 8C:60:4F:73:2D:57 |   0   | False  | 222  |
            ______________________________________________________________________________
            |  True  |  ae1.222  |    0.0    | F4:0F:1B:76:9D:97 |   0   | False  | 222  |
            ______________________________________________________________________________
        result:
        ----------


CLI Example:

.. code-block:: bash

    salt '*' foo.bar --out=table
"""

import operator
from functools import reduce  # pylint: disable=redefined-builtin

import salt.output
import salt.utils.color
import salt.utils.data

__virtualname__ = "table"


def __virtual__():
    return __virtualname__


class TableDisplay:
    """
    Manage the table display content.
    """

    _JUSTIFY_MAP = {
        "center": str.center,
        "right": str.rjust,
        "left": str.ljust,
    }

    def __init__(
        self,
        has_header=True,  # if header will be displayed
        row_delimiter="-",  # row delimiter char
        delim=" | ",  # column delimiter
        justify="center",  # text justify
        separate_rows=True,  # display the line separating two consecutive rows
        prefix="| ",  # character to display at the beginning of the row
        suffix=" |",  # character to display at the end of the row
        width=50,  # column max width
        wrapfunc=None,
    ):  # function wrapper
        self.__dict__.update(
            salt.utils.color.get_colors(
                __opts__.get("color"), __opts__.get("color_theme")
            )
        )
        self.strip_colors = __opts__.get("strip_colors", True)

        self.has_header = has_header
        self.row_delimiter = row_delimiter
        self.delim = delim
        self.justify = justify
        self.separate_rows = separate_rows
        self.prefix = prefix
        self.suffix = suffix
        self.width = width

        if not (wrapfunc and callable(wrapfunc)):
            self.wrapfunc = self.wrap_onspace
        else:
            self.wrapfunc = wrapfunc

    def ustring(self, indent, color, msg, prefix="", suffix="", endc=None):
        """Build the unicode string to be displayed."""
        if endc is None:
            endc = self.ENDC  # pylint: disable=no-member

        indent *= " "
        fmt = "{0}{1}{2}{3}{4}{5}"

        try:
            return fmt.format(indent, color, prefix, msg, endc, suffix)
        except UnicodeDecodeError:
            return fmt.format(
                indent, color, prefix, salt.utils.data.decode(msg), endc, suffix
            )

    def wrap_onspace(self, text):
        """
        When the text inside the column is longer then the width, will split by space and continue on the next line.
        """

        def _truncate(line, word):
            return "{line}{part}{word}".format(
                line=line,
                part=" \n"[
                    (
                        len(line[line.rfind("\n") + 1 :]) + len(word.split("\n", 1)[0])
                        >= self.width
                    )
                ],
                word=word,
            )

        return reduce(_truncate, text.split(" "))

    def prepare_rows(self, rows, indent, has_header):
        """Prepare rows content to be displayed."""

        out = []

        def row_wrapper(row):
            new_rows = [self.wrapfunc(item).split("\n") for item in row]
            rows = []
            for item in map(lambda *args: args, *new_rows):
                if isinstance(item, (tuple, list)):
                    rows.append([substr or "" for substr in item])
                else:
                    rows.append([item])
            return rows

        logical_rows = [row_wrapper(row) for row in rows]

        columns = map(lambda *args: args, *reduce(operator.add, logical_rows))

        max_widths = [max(len(str(item)) for item in column) for column in columns]
        row_separator = self.row_delimiter * (
            len(self.prefix)
            + len(self.suffix)
            + sum(max_widths)
            + len(self.delim) * (len(max_widths) - 1)
        )

        justify = self._JUSTIFY_MAP[self.justify.lower()]

        if self.separate_rows:
            out.append(
                self.ustring(
                    indent, self.LIGHT_GRAY, row_separator  # pylint: disable=no-member
                )
            )
        for physical_rows in logical_rows:
            for row in physical_rows:
                line = (
                    self.prefix
                    + self.delim.join(
                        [
                            justify(str(item), width)
                            for (item, width) in zip(row, max_widths)
                        ]
                    )
                    + self.suffix
                )
                out.append(
                    self.ustring(indent, self.WHITE, line)  # pylint: disable=no-member
                )
            if self.separate_rows or has_header:
                out.append(
                    self.ustring(
                        indent,
                        self.LIGHT_GRAY,  # pylint: disable=no-member
                        row_separator,
                    )
                )
                has_header = False
        return out

    def display_rows(self, rows, labels, indent):
        """Prepares row content and displays."""

        out = []

        if not rows:
            return out

        first_row_type = type(rows[0])
        # all rows must have the same datatype
        consistent = True
        for row in rows[1:]:
            if type(row) != first_row_type:
                consistent = False

        if not consistent:
            return out

        if isinstance(labels, dict):
            labels_temp = []
            for key in sorted(labels):
                labels_temp.append(labels[key])
            labels = labels_temp

        if first_row_type is dict:  # and all the others
            temp_rows = []
            if not labels:
                labels = [
                    str(label).replace("_", " ").title() for label in sorted(rows[0])
                ]
            for row in rows:
                temp_row = []
                for key in sorted(row):
                    temp_row.append(str(row[key]))
                temp_rows.append(temp_row)
            rows = temp_rows
        elif isinstance(rows[0], str):
            rows = [
                [row] for row in rows
            ]  # encapsulate each row in a single-element list

        labels_and_rows = [labels] + rows if labels else rows
        has_header = self.has_header and labels

        return self.prepare_rows(labels_and_rows, indent + 4, has_header)

    def display(self, ret, indent, out, rows_key=None, labels_key=None):
        """Display table(s)."""

        rows = []
        labels = None

        if isinstance(ret, dict):
            if not rows_key or (rows_key and rows_key in list(ret.keys())):
                # either not looking for a specific key
                # either looking and found in the current root
                for key in sorted(ret):
                    if rows_key and key != rows_key:
                        continue  # if searching specifics, ignore anything else
                    val = ret[key]
                    if not rows_key:
                        out.append(
                            self.ustring(
                                indent,
                                self.DARK_GRAY,  # pylint: disable=no-member
                                key,
                                suffix=":",
                            )
                        )
                        out.append(
                            self.ustring(
                                indent,
                                self.DARK_GRAY,  # pylint: disable=no-member
                                "----------",
                            )
                        )
                    if isinstance(val, (list, tuple)):
                        rows = val
                        if labels_key:
                            # at the same depth
                            labels = ret.get(labels_key)  # if any
                        out.extend(self.display_rows(rows, labels, indent))
                    else:
                        self.display(
                            val,
                            indent + 4,
                            out,
                            rows_key=rows_key,
                            labels_key=labels_key,
                        )
            elif rows_key:
                # dig deeper
                for key in sorted(ret):
                    val = ret[key]
                    self.display(
                        val, indent, out, rows_key=rows_key, labels_key=labels_key
                    )  # same indent
        elif isinstance(ret, (list, tuple)):
            if not rows_key:
                rows = ret
                out.extend(self.display_rows(rows, labels, indent))

        return out


def output(ret, **kwargs):
    """
    Display the output as table.

    Args:

        * nested_indent: integer, specify the left alignment.
        * has_header: boolean specifying if header should be displayed. Default: True.
        * row_delimiter: character to separate rows. Default: ``_``.
        * delim: character to separate columns. Default: ``" | "``.
        * justify: text alignment. Default: ``center``.
        * separate_rows: boolean specifying if row separator will be displayed between consecutive rows. Default: True.
        * prefix: character at the beginning of the row. Default: ``"| "``.
        * suffix: character at the end of the row. Default: ``" |"``.
        * width: column max width. Default: ``50``.
        * rows_key: display the rows under a specific key.
        * labels_key: use the labels under a certain key. Otherwise will try to use the dictionary keys (if any).
        * title: display title when only one table is selected (using the ``rows_key`` argument).
    """

    # to facilitate re-use
    if "opts" in kwargs:
        global __opts__  # pylint: disable=W0601
        __opts__ = kwargs.pop("opts")

    # Prefer kwargs before opts
    base_indent = kwargs.get("nested_indent", 0) or __opts__.get(
        "out.table.nested_indent", 0
    )
    rows_key = kwargs.get("rows_key") or __opts__.get("out.table.rows_key")
    labels_key = kwargs.get("labels_key") or __opts__.get("out.table.labels_key")
    title = kwargs.get("title") or __opts__.get("out.table.title")

    class_kvargs = {}
    argks = (
        "has_header",
        "row_delimiter",
        "delim",
        "justify",
        "separate_rows",
        "prefix",
        "suffix",
        "width",
    )

    for argk in argks:
        argv = kwargs.get(argk) or __opts__.get(f"out.table.{argk}")
        if argv is not None:
            class_kvargs[argk] = argv

    table = TableDisplay(**class_kvargs)

    out = []
    if title and rows_key:
        out.append(
            table.ustring(
                base_indent,
                title,
                table.WHITE,  # pylint: disable=no-member
                suffix="\n",
            )
        )

    return "\n".join(
        table.display(
            salt.utils.data.decode(ret),
            base_indent,
            out,
            rows_key=rows_key,
            labels_key=labels_key,
        )
    )



"""
Integration tests for the lxd states
"""

import pytest

import salt.modules.lxd
from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin


@pytest.mark.destructive_test
@pytest.mark.skipif(salt.modules.lxd.HAS_PYLXD is False, reason="pylxd not installed")
@pytest.mark.skip_if_binaries_missing("lxd", reason="LXD not installed")
@pytest.mark.skip_if_binaries_missing("lxc", reason="LXC not installed")
@pytest.mark.slow_test
class LxdTestCase(ModuleCase, SaltReturnAssertsMixin):
    @pytest.mark.flaky(max_runs=4)
    def test_01__init_lxd(self):
        ret = self.run_state("lxd.init", name="foobar")
        self.assertSaltTrueReturn(ret)
        name = "lxd_|-foobar_|-foobar_|-init"
        assert name in ret
        assert ret[name]["storage_backend"] == "dir"



"""
AliYun ECS Cloud Module
=======================

.. versionadded:: 2014.7.0

The Aliyun cloud module is used to control access to the aliyun ECS.
http://www.aliyun.com/

Use of this module requires the ``id`` and ``key`` parameter to be set.
Set up the cloud configuration at ``/etc/salt/cloud.providers`` or
``/etc/salt/cloud.providers.d/aliyun.conf``:

.. code-block:: yaml

    my-aliyun-config:
      # aliyun Access Key ID
      id: wFGEwgregeqw3435gDger
      # aliyun Access Key Secret
      key: GDE43t43REGTrkilg43934t34qT43t4dgegerGEgg
      location: cn-qingdao
      driver: aliyun

:depends: requests
"""

import base64
import hmac
import logging
import pprint
import sys
import time
import urllib.parse
import uuid
from hashlib import sha1

import salt.config as config
import salt.utils.cloud
import salt.utils.data
import salt.utils.json
from salt.exceptions import (
    SaltCloudExecutionFailure,
    SaltCloudExecutionTimeout,
    SaltCloudNotFound,
    SaltCloudSystemExit,
)
from salt.utils.stringutils import to_bytes

try:
    import requests

    HAS_REQUESTS = True
except ImportError:
    HAS_REQUESTS = False

# Get logging started
log = logging.getLogger(__name__)

ALIYUN_LOCATIONS = {
    # 'us-west-2': 'ec2_us_west_oregon',
    "cn-hangzhou": "AliYun HangZhou Region",
    "cn-beijing": "AliYun BeiJing Region",
    "cn-hongkong": "AliYun HongKong Region",
    "cn-qingdao": "AliYun QingDao Region",
    "cn-shanghai": "AliYun ShangHai Region",
    "cn-shenzhen": "AliYun ShenZheng Region",
    "ap-northeast-1": "AliYun DongJing Region",
    "ap-southeast-1": "AliYun XinJiaPo Region",
    "ap-southeast-2": "AliYun XiNi Region",
    "eu-central-1": "EU FalaKeFu Region",
    "me-east-1": "ME DiBai Region",
    "us-east-1": "US FuJiNiYa Region",
    "us-west-1": "US GuiGu Region",
}
DEFAULT_LOCATION = "cn-hangzhou"

DEFAULT_ALIYUN_API_VERSION = "2014-05-26"

__virtualname__ = "aliyun"


# Only load in this module if the aliyun configurations are in place
def __virtual__():
    """
    Check for aliyun configurations
    """
    if get_configured_provider() is False:
        return False

    if get_dependencies() is False:
        return False

    return __virtualname__


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def get_configured_provider():
    """
    Return the first configured instance.
    """
    return config.is_provider_configured(
        __opts__, _get_active_provider_name() or __virtualname__, ("id", "key")
    )


def get_dependencies():
    """
    Warn if dependencies aren't met.
    """
    return config.check_driver_dependencies(__virtualname__, {"requests": HAS_REQUESTS})


def avail_locations(call=None):
    """
    Return a dict of all available VM locations on the cloud provider with
    relevant data
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_locations function must be called with "
            "-f or --function, or with the --list-locations option"
        )

    params = {"Action": "DescribeRegions"}
    items = query(params=params)

    ret = {}
    for region in items["Regions"]["Region"]:
        ret[region["RegionId"]] = {}
        for item in region:
            ret[region["RegionId"]][item] = str(region[item])

    return ret


def avail_images(kwargs=None, call=None):
    """
    Return a list of the images that are on the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_images function must be called with "
            "-f or --function, or with the --list-images option"
        )

    if not isinstance(kwargs, dict):
        kwargs = {}

    provider = get_configured_provider()
    location = provider.get("location", DEFAULT_LOCATION)

    if "location" in kwargs:
        location = kwargs["location"]

    params = {
        "Action": "DescribeImages",
        "RegionId": location,
        "PageSize": "100",
    }
    items = query(params=params)

    ret = {}
    for image in items["Images"]["Image"]:
        ret[image["ImageId"]] = {}
        for item in image:
            ret[image["ImageId"]][item] = str(image[item])

    return ret


def avail_sizes(call=None):
    """
    Return a list of the image sizes that are on the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_sizes function must be called with "
            "-f or --function, or with the --list-sizes option"
        )

    params = {"Action": "DescribeInstanceTypes"}
    items = query(params=params)

    ret = {}
    for image in items["InstanceTypes"]["InstanceType"]:
        ret[image["InstanceTypeId"]] = {}
        for item in image:
            ret[image["InstanceTypeId"]][item] = str(image[item])

    return ret


def get_location(vm_=None):
    """
    Return the aliyun region to use, in this order:
        - CLI parameter
        - VM parameter
        - Cloud profile setting
    """
    return __opts__.get(
        "location",
        config.get_cloud_config_value(
            "location",
            vm_ or get_configured_provider(),
            __opts__,
            default=DEFAULT_LOCATION,
            search_global=False,
        ),
    )


def list_availability_zones(call=None):
    """
    List all availability zones in the current region
    """
    ret = {}

    params = {"Action": "DescribeZones", "RegionId": get_location()}
    items = query(params)

    for zone in items["Zones"]["Zone"]:
        ret[zone["ZoneId"]] = {}
        for item in zone:
            ret[zone["ZoneId"]][item] = str(zone[item])

    return ret


def list_nodes_min(call=None):
    """
    Return a list of the VMs that are on the provider. Only a list of VM names,
    and their state, is returned. This is the minimum amount of information
    needed to check for existing VMs.
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes_min function must be called with -f or --function."
        )

    ret = {}
    location = get_location()
    params = {
        "Action": "DescribeInstanceStatus",
        "RegionId": location,
    }
    nodes = query(params)

    log.debug("Total %s instance found in Region %s", nodes["TotalCount"], location)
    if "Code" in nodes or nodes["TotalCount"] == 0:
        return ret

    for node in nodes["InstanceStatuses"]["InstanceStatus"]:
        ret[node["InstanceId"]] = {}
        for item in node:
            ret[node["InstanceId"]][item] = node[item]

    return ret


def list_nodes(call=None):
    """
    Return a list of the VMs that are on the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes function must be called with -f or --function."
        )

    nodes = list_nodes_full()
    ret = {}
    for instanceId in nodes:
        node = nodes[instanceId]
        ret[node["name"]] = {
            "id": node["id"],
            "name": node["name"],
            "public_ips": node["public_ips"],
            "private_ips": node["private_ips"],
            "size": node["size"],
            "state": str(node["state"]),
        }
    return ret


def list_nodes_full(call=None):
    """
    Return a list of the VMs that are on the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes_full function must be called with -f or --function."
        )

    ret = {}
    location = get_location()
    params = {
        "Action": "DescribeInstanceStatus",
        "RegionId": location,
        "PageSize": "50",
    }
    result = query(params=params)

    log.debug("Total %s instance found in Region %s", result["TotalCount"], location)
    if "Code" in result or result["TotalCount"] == 0:
        return ret

    # aliyun max 100 top instance in api
    result_instancestatus = result["InstanceStatuses"]["InstanceStatus"]
    if result["TotalCount"] > 50:
        params["PageNumber"] = "2"
        result = query(params=params)
        result_instancestatus.update(result["InstanceStatuses"]["InstanceStatus"])

    for node in result_instancestatus:

        instanceId = node.get("InstanceId", "")

        params = {"Action": "DescribeInstanceAttribute", "InstanceId": instanceId}
        items = query(params=params)
        if "Code" in items:
            log.warning("Query instance:%s attribute failed", instanceId)
            continue

        name = items["InstanceName"]
        ret[name] = {
            "id": items["InstanceId"],
            "name": name,
            "image": items["ImageId"],
            "size": "TODO",
            "state": items["Status"],
        }
        for item in items:
            value = items[item]
            if value is not None:
                value = str(value)
            if item == "PublicIpAddress":
                ret[name]["public_ips"] = items[item]["IpAddress"]
            if item == "InnerIpAddress" and "private_ips" not in ret[name]:
                ret[name]["private_ips"] = items[item]["IpAddress"]
            if item == "VpcAttributes":
                vpc_ips = items[item]["PrivateIpAddress"]["IpAddress"]
                if vpc_ips:
                    ret[name]["private_ips"] = vpc_ips
            ret[name][item] = value

    provider = _get_active_provider_name() or "aliyun"
    if ":" in provider:
        comps = provider.split(":")
        provider = comps[0]

    __opts__["update_cachedir"] = True
    __utils__["cloud.cache_node_list"](ret, provider, __opts__)

    return ret


def list_nodes_select(call=None):
    """
    Return a list of the VMs that are on the provider, with select fields
    """
    return salt.utils.cloud.list_nodes_select(
        list_nodes_full("function"),
        __opts__["query.selection"],
        call,
    )


def list_securitygroup(call=None):
    """
    Return a list of security group
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes function must be called with -f or --function."
        )

    params = {
        "Action": "DescribeSecurityGroups",
        "RegionId": get_location(),
        "PageSize": "50",
    }

    result = query(params)
    if "Code" in result:
        return {}

    ret = {}
    for sg in result["SecurityGroups"]["SecurityGroup"]:
        ret[sg["SecurityGroupId"]] = {}
        for item in sg:
            ret[sg["SecurityGroupId"]][item] = sg[item]

    return ret


def get_image(vm_):
    """
    Return the image object to use
    """
    images = avail_images()
    vm_image = str(
        config.get_cloud_config_value("image", vm_, __opts__, search_global=False)
    )

    if not vm_image:
        raise SaltCloudNotFound("No image specified for this VM.")

    if vm_image and str(vm_image) in images:
        return images[vm_image]["ImageId"]
    raise SaltCloudNotFound(f"The specified image, '{vm_image}', could not be found.")


def get_securitygroup(vm_):
    """
    Return the security group
    """
    sgs = list_securitygroup()
    securitygroup = config.get_cloud_config_value(
        "securitygroup", vm_, __opts__, search_global=False
    )

    if not securitygroup:
        raise SaltCloudNotFound("No securitygroup ID specified for this VM.")

    if securitygroup and str(securitygroup) in sgs:
        return sgs[securitygroup]["SecurityGroupId"]
    raise SaltCloudNotFound(
        f"The specified security group, '{securitygroup}', could not be found."
    )


def get_size(vm_):
    """
    Return the VM's size. Used by create_node().
    """
    sizes = avail_sizes()
    vm_size = str(
        config.get_cloud_config_value("size", vm_, __opts__, search_global=False)
    )

    if not vm_size:
        raise SaltCloudNotFound("No size specified for this VM.")

    if vm_size and str(vm_size) in sizes:
        return sizes[vm_size]["InstanceTypeId"]

    raise SaltCloudNotFound(f"The specified size, '{vm_size}', could not be found.")


def __get_location(vm_):
    """
    Return the VM's location
    """
    locations = avail_locations()
    vm_location = str(
        config.get_cloud_config_value("location", vm_, __opts__, search_global=False)
    )

    if not vm_location:
        raise SaltCloudNotFound("No location specified for this VM.")

    if vm_location and str(vm_location) in locations:
        return locations[vm_location]["RegionId"]
    raise SaltCloudNotFound(
        f"The specified location, '{vm_location}', could not be found."
    )


def start(name, call=None):
    """
    Start a node

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a start myinstance
    """
    if call != "action":
        raise SaltCloudSystemExit("The stop action must be called with -a or --action.")

    log.info("Starting node %s", name)

    instanceId = _get_node(name)["InstanceId"]

    params = {"Action": "StartInstance", "InstanceId": instanceId}
    result = query(params)

    return result


def stop(name, force=False, call=None):
    """
    Stop a node

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a stop myinstance
        salt-cloud -a stop myinstance force=True
    """
    if call != "action":
        raise SaltCloudSystemExit("The stop action must be called with -a or --action.")

    log.info("Stopping node %s", name)

    instanceId = _get_node(name)["InstanceId"]

    params = {
        "Action": "StopInstance",
        "InstanceId": instanceId,
        "ForceStop": str(force).lower(),
    }
    result = query(params)

    return result


def reboot(name, call=None):
    """
    Reboot a node

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a reboot myinstance
    """
    if call != "action":
        raise SaltCloudSystemExit("The stop action must be called with -a or --action.")

    log.info("Rebooting node %s", name)

    instance_id = _get_node(name)["InstanceId"]

    params = {"Action": "RebootInstance", "InstanceId": instance_id}
    result = query(params)

    return result


def create_node(kwargs):
    """
    Convenience function to make the rest api call for node creation.
    """
    if not isinstance(kwargs, dict):
        kwargs = {}

    # Required parameters
    params = {
        "Action": "CreateInstance",
        "InstanceType": kwargs.get("size_id", ""),
        "RegionId": kwargs.get("region_id", DEFAULT_LOCATION),
        "ImageId": kwargs.get("image_id", ""),
        "SecurityGroupId": kwargs.get("securitygroup_id", ""),
        "InstanceName": kwargs.get("name", ""),
    }

    # Optional parameters'
    optional = [
        "InstanceName",
        "InternetChargeType",
        "InternetMaxBandwidthIn",
        "InternetMaxBandwidthOut",
        "HostName",
        "Password",
        "SystemDisk.Category",
        "VSwitchId",
        # 'DataDisk.n.Size', 'DataDisk.n.Category', 'DataDisk.n.SnapshotId'
    ]

    for item in optional:
        if item in kwargs:
            params.update({item: kwargs[item]})

    # invoke web call
    result = query(params)
    return result["InstanceId"]


def create(vm_):
    """
    Create a single VM from a data dict
    """
    try:
        # Check for required profile parameters before sending any API calls.
        if (
            vm_["profile"]
            and config.is_profile_configured(
                __opts__,
                _get_active_provider_name() or "aliyun",
                vm_["profile"],
                vm_=vm_,
            )
            is False
        ):
            return False
    except AttributeError:
        pass

    __utils__["cloud.fire_event"](
        "event",
        "starting create",
        "salt/cloud/{}/creating".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "creating", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    log.info("Creating Cloud VM %s", vm_["name"])
    kwargs = {
        "name": vm_["name"],
        "size_id": get_size(vm_),
        "image_id": get_image(vm_),
        "region_id": __get_location(vm_),
        "securitygroup_id": get_securitygroup(vm_),
    }
    if "vswitch_id" in vm_:
        kwargs["VSwitchId"] = vm_["vswitch_id"]
    if "internet_chargetype" in vm_:
        kwargs["InternetChargeType"] = vm_["internet_chargetype"]
    if "internet_maxbandwidthin" in vm_:
        kwargs["InternetMaxBandwidthIn"] = str(vm_["internet_maxbandwidthin"])
    if "internet_maxbandwidthout" in vm_:
        kwargs["InternetMaxBandwidthOut"] = str(vm_["internet_maxbandwidthOut"])
    if "hostname" in vm_:
        kwargs["HostName"] = vm_["hostname"]
    if "password" in vm_:
        kwargs["Password"] = vm_["password"]
    if "instance_name" in vm_:
        kwargs["InstanceName"] = vm_["instance_name"]
    if "systemdisk_category" in vm_:
        kwargs["SystemDisk.Category"] = vm_["systemdisk_category"]

    __utils__["cloud.fire_event"](
        "event",
        "requesting instance",
        "salt/cloud/{}/requesting".format(vm_["name"]),
        args=__utils__["cloud.filter_event"]("requesting", kwargs, list(kwargs)),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    try:
        ret = create_node(kwargs)
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            "Error creating %s on Aliyun ECS\n\n"
            "The following exception was thrown when trying to "
            "run the initial deployment: %s",
            vm_["name"],
            str(exc),
            # Show the traceback if the debug logging level is enabled
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False
    # repair ip address error and start vm
    time.sleep(8)
    params = {"Action": "StartInstance", "InstanceId": ret}
    query(params)

    def __query_node_data(vm_name):
        data = show_instance(vm_name, call="action")
        if not data:
            # Trigger an error in the wait_for_ip function
            return False
        if data.get("PublicIpAddress", None) is not None:
            return data

    try:
        data = salt.utils.cloud.wait_for_ip(
            __query_node_data,
            update_args=(vm_["name"],),
            timeout=config.get_cloud_config_value(
                "wait_for_ip_timeout", vm_, __opts__, default=10 * 60
            ),
            interval=config.get_cloud_config_value(
                "wait_for_ip_interval", vm_, __opts__, default=10
            ),
        )
    except (SaltCloudExecutionTimeout, SaltCloudExecutionFailure) as exc:
        try:
            # It might be already up, let's destroy it!
            destroy(vm_["name"])
        except SaltCloudSystemExit:
            pass
        finally:
            raise SaltCloudSystemExit(str(exc))

    if data["public_ips"]:
        ssh_ip = data["public_ips"][0]
    elif data["private_ips"]:
        ssh_ip = data["private_ips"][0]
    else:
        log.info("No available ip:cant connect to salt")
        return False
    log.debug("VM %s is now running", ssh_ip)
    vm_["ssh_host"] = ssh_ip

    # The instance is booted and accessible, let's Salt it!
    ret = __utils__["cloud.bootstrap"](vm_, __opts__)
    ret.update(data)

    log.info("Created Cloud VM '%s'", vm_["name"])
    log.debug("'%s' VM creation details:\n%s", vm_["name"], pprint.pformat(data))

    __utils__["cloud.fire_event"](
        "event",
        "created instance",
        "salt/cloud/{}/created".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "created", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return ret


def _compute_signature(parameters, access_key_secret):
    """
    Generate aliyun request signature
    """

    def percent_encode(line):
        if not isinstance(line, str):
            return line

        s = line
        if sys.stdin.encoding is None:
            s = line.decode().encode("utf8")
        else:
            s = line.decode(sys.stdin.encoding).encode("utf8")
        res = urllib.parse.quote(s, "")
        res = res.replace("+", "%20")
        res = res.replace("*", "%2A")
        res = res.replace("%7E", "~")
        return res

    sortedParameters = sorted(list(parameters.items()), key=lambda items: items[0])

    canonicalizedQueryString = ""
    for k, v in sortedParameters:
        canonicalizedQueryString += "&" + percent_encode(k) + "=" + percent_encode(v)

    # All aliyun API only support GET method
    stringToSign = "GET&%2F&" + percent_encode(canonicalizedQueryString[1:])

    h = hmac.new(to_bytes(access_key_secret + "&"), stringToSign, sha1)
    signature = base64.encodestring(h.digest()).strip()
    return signature


def query(params=None):
    """
    Make a web call to aliyun ECS REST API
    """
    path = "https://ecs-cn-hangzhou.aliyuncs.com"

    access_key_id = config.get_cloud_config_value(
        "id", get_configured_provider(), __opts__, search_global=False
    )
    access_key_secret = config.get_cloud_config_value(
        "key", get_configured_provider(), __opts__, search_global=False
    )

    timestamp = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())

    # public interface parameters
    parameters = {
        "Format": "JSON",
        "Version": DEFAULT_ALIYUN_API_VERSION,
        "AccessKeyId": access_key_id,
        "SignatureVersion": "1.0",
        "SignatureMethod": "HMAC-SHA1",
        "SignatureNonce": str(uuid.uuid1()),
        "TimeStamp": timestamp,
    }

    # include action or function parameters
    if params:
        parameters.update(params)

    # Calculate the string for Signature
    signature = _compute_signature(parameters, access_key_secret)
    parameters["Signature"] = signature

    request = requests.get(path, params=parameters, verify=True, timeout=120)
    if request.status_code != 200:
        raise SaltCloudSystemExit(
            "An error occurred while querying aliyun ECS. HTTP Code: {}  "
            "Error: '{}'".format(request.status_code, request.text)
        )

    log.debug(request.url)

    content = request.text

    result = salt.utils.json.loads(content)
    if "Code" in result:
        raise SaltCloudSystemExit(pprint.pformat(result.get("Message", {})))
    return result


def script(vm_):
    """
    Return the script deployment object
    """
    deploy_script = salt.utils.cloud.os_script(
        config.get_cloud_config_value("script", vm_, __opts__),
        vm_,
        __opts__,
        salt.utils.cloud.salt_config_to_yaml(
            salt.utils.cloud.minion_config(__opts__, vm_)
        ),
    )
    return deploy_script


def show_disk(name, call=None):
    """
    Show the disk details of the instance

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a show_disk aliyun myinstance
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The show_disks action must be called with -a or --action."
        )

    ret = {}
    params = {"Action": "DescribeInstanceDisks", "InstanceId": name}
    items = query(params=params)

    for disk in items["Disks"]["Disk"]:
        ret[disk["DiskId"]] = {}
        for item in disk:
            ret[disk["DiskId"]][item] = str(disk[item])

    return ret


def list_monitor_data(kwargs=None, call=None):
    """
    Get monitor data of the instance. If instance name is
    missing, will show all the instance monitor data on the region.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -f list_monitor_data aliyun
        salt-cloud -f list_monitor_data aliyun name=AY14051311071990225bd
    """
    if call != "function":
        raise SaltCloudSystemExit(
            "The list_monitor_data must be called with -f or --function."
        )

    if not isinstance(kwargs, dict):
        kwargs = {}

    ret = {}
    params = {"Action": "GetMonitorData", "RegionId": get_location()}
    if "name" in kwargs:
        params["InstanceId"] = kwargs["name"]

    items = query(params=params)

    monitorData = items["MonitorData"]

    for data in monitorData["InstanceMonitorData"]:
        ret[data["InstanceId"]] = {}
        for item in data:
            ret[data["InstanceId"]][item] = str(data[item])

    return ret


def show_instance(name, call=None):
    """
    Show the details from aliyun instance
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The show_instance action must be called with -a or --action."
        )

    return _get_node(name)


def _get_node(name):
    attempts = 5
    while attempts >= 0:
        try:
            return list_nodes_full()[name]
        except KeyError:
            attempts -= 1
            log.debug(
                "Failed to get the data for node '%s'. Remaining attempts: %s",
                name,
                attempts,
            )
            # Just a little delay between attempts...
            time.sleep(0.5)
    raise SaltCloudNotFound(f"The specified instance {name} not found")


def show_image(kwargs, call=None):
    """
    Show the details from aliyun image
    """
    if call != "function":
        raise SaltCloudSystemExit(
            "The show_images function must be called with -f or --function"
        )

    if not isinstance(kwargs, dict):
        kwargs = {}

    location = get_location()
    if "location" in kwargs:
        location = kwargs["location"]

    params = {
        "Action": "DescribeImages",
        "RegionId": location,
        "ImageId": kwargs["image"],
    }

    ret = {}
    items = query(params=params)
    # DescribeImages so far support input multi-image. And
    # if not found certain image, the response will include
    # blank image list other than 'not found' error message
    if "Code" in items or not items["Images"]["Image"]:
        raise SaltCloudNotFound("The specified image could not be found.")

    log.debug("Total %s image found in Region %s", items["TotalCount"], location)

    for image in items["Images"]["Image"]:
        ret[image["ImageId"]] = {}
        for item in image:
            ret[image["ImageId"]][item] = str(image[item])

    return ret


def destroy(name, call=None):
    """
    Destroy a node.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a destroy myinstance
        salt-cloud -d myinstance
    """
    if call == "function":
        raise SaltCloudSystemExit(
            "The destroy action must be called with -d, --destroy, -a or --action."
        )

    __utils__["cloud.fire_event"](
        "event",
        "destroying instance",
        f"salt/cloud/{name}/destroying",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    instanceId = _get_node(name)["InstanceId"]

    # have to stop instance before del it
    stop_params = {"Action": "StopInstance", "InstanceId": instanceId}
    query(stop_params)

    params = {"Action": "DeleteInstance", "InstanceId": instanceId}

    node = query(params)

    __utils__["cloud.fire_event"](
        "event",
        "destroyed instance",
        f"salt/cloud/{name}/destroyed",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return node



import datetime
import logging

import pytest

try:
    import dateutil.parser

    HAS_DATEUTIL_PARSER = True
except ImportError:
    HAS_DATEUTIL_PARSER = False

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skipif(
        HAS_DATEUTIL_PARSER is False,
        reason="The 'dateutil.parser' library is not available",
    ),
    pytest.mark.windows_whitelisted,
]


@pytest.mark.slow_test
def test_postpone(schedule):
    """
    verify that scheduled job is postponed until the specified time.
    """
    job = {"schedule": {"job1": {"function": "test.ping", "when": "11/29/2017 4pm"}}}

    # 11/29/2017 4pm
    run_time = dateutil.parser.parse("11/29/2017 4:00pm")

    # 5 minute delay
    delay = 300

    # Add job to schedule
    schedule.opts.update(job)

    # Postpone the job by 5 minutes
    schedule.postpone_job(
        "job1",
        {
            "time": run_time.strftime("%Y-%m-%dT%H:%M:%S"),
            "new_time": (run_time + datetime.timedelta(seconds=delay)).strftime(
                "%Y-%m-%dT%H:%M:%S"
            ),
        },
    )
    # Run at the original time
    schedule.eval(now=run_time)
    ret = schedule.job_status("job1")
    assert "_last_run" not in ret

    # Run 5 minutes later
    schedule.eval(now=run_time + datetime.timedelta(seconds=delay))
    ret = schedule.job_status("job1")
    assert ret["_last_run"] == run_time + datetime.timedelta(seconds=delay)

    # Run 6 minutes later
    schedule.eval(now=run_time + datetime.timedelta(seconds=delay + 1))
    ret = schedule.job_status("job1")
    assert ret["_last_run"] == run_time + datetime.timedelta(seconds=delay)



import pytest

from tests.support.case import ModuleCase


@pytest.mark.skip_unless_on_windows
@pytest.mark.windows_whitelisted
class FirewallTest(ModuleCase):
    """
    Validate windows firewall module
    """

    def _pre_firewall_status(self, pre_run):
        post_run = self.run_function("firewall.get_config")
        network = ["Domain", "Public", "Private"]
        # compare the status of the firewall before and after test
        # and re-enable or disable depending on status before test run
        for net in network:
            if post_run[net] != pre_run[net]:
                if pre_run[net]:
                    self.assertTrue(self.run_function("firewall.enable", profile=net))
                else:
                    self.assertTrue(self.run_function("firewall.disable", profile=net))

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_firewall_get_config(self):
        """
        test firewall.get_config
        """
        pre_run = self.run_function("firewall.get_config")
        # ensure all networks are enabled then test status
        self.assertTrue(self.run_function("firewall.enable", profile="allprofiles"))
        ret = self.run_function("firewall.get_config")
        network = ["Domain", "Public", "Private"]
        for net in network:
            self.assertTrue(ret[net])
        self._pre_firewall_status(pre_run)

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_firewall_disable(self):
        """
        test firewall.disable
        """
        pre_run = self.run_function("firewall.get_config")
        network = "Private"

        ret = self.run_function("firewall.get_config")[network]
        if not ret:
            self.assertTrue(self.run_function("firewall.enable", profile=network))

        self.assertTrue(self.run_function("firewall.disable", profile=network))
        ret = self.run_function("firewall.get_config")[network]
        self.assertFalse(ret)
        self._pre_firewall_status(pre_run)

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_firewall_enable(self):
        """
        test firewall.enable
        """
        pre_run = self.run_function("firewall.get_config")
        network = "Private"

        ret = self.run_function("firewall.get_config")[network]
        if ret:
            self.assertTrue(self.run_function("firewall.disable", profile=network))

        self.assertTrue(self.run_function("firewall.enable", profile=network))
        ret = self.run_function("firewall.get_config")[network]
        self.assertTrue(ret)
        self._pre_firewall_status(pre_run)

    @pytest.mark.slow_test
    def test_firewall_get_rule(self):
        """
        test firewall.get_rule
        """
        rule = "Remote Event Log Management (NP-In)"

        ret = self.run_function("firewall.get_rule", [rule])
        checks = ["Private", "LocalPort", "RemotePort"]
        for check in checks:
            self.assertIn(check, ret[rule])

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_firewall_add_delete_rule(self):
        """
        test firewall.add_rule and delete_rule
        """
        rule = "test rule"
        port = "8080"

        # test adding firewall rule
        add_rule = self.run_function("firewall.add_rule", [rule, port])
        ret = self.run_function("firewall.get_rule", [rule])
        self.assertIn(rule, ret[rule])
        self.assertIn(port, ret[rule])

        # test deleting firewall rule
        self.assertTrue(self.run_function("firewall.delete_rule", [rule, port]))
        ret = self.run_function("firewall.get_rule", [rule])
        self.assertNotIn(rule, ret)
        self.assertNotIn(port, ret)
        self.assertIn("No rules match the specified criteria.", ret)



import os

import pytest

import salt.scripts
import salt.utils.platform
from tests.conftest import CODE_DIR
from tests.support.mock import patch


def test_within_onedir_env(shell):
    if os.environ.get("ONEDIR_TESTRUN", "0") == "0":
        return

    script_name = "salt-pip"
    if salt.utils.platform.is_windows():
        script_name += ".exe"

    script_path = CODE_DIR / "artifacts" / "salt" / script_name
    assert script_path.exists()

    ret = shell.run(str(script_path), "list")
    assert ret.returncode == 0


def test_outside_onedir_env(capsys):
    with patch("salt.scripts._get_onedir_env_path", return_value=None):
        with pytest.raises(SystemExit) as exc:
            salt.scripts.salt_pip()
    captured = capsys.readouterr()
    assert "'salt-pip' is only meant to be used from a Salt onedir." in captured.err



"""
Utility functions for the rest_sample
"""

__proxyenabled__ = ["rest_sample"]


def fix_outage():
    """
    "Fix" the outage

    CLI Example:

    .. code-block:: bash

        salt 'rest-sample-proxy' rest_sample.fix_outage

    """
    return __proxy__["rest_sample.fix_outage"]()


def get_test_string():
    """
    Helper function to test cross-calling to the __proxy__ dunder.

    CLI Example:

    .. code-block:: bash

        salt 'rest-sample-proxy' rest_sample.get_test_string
    """
    return __proxy__["rest_sample.test_from_state"]()



import time

import pytest

pytestmark = [
    pytest.mark.skip_unless_on_linux(reason="Only supported on Linux family"),
    pytest.mark.skipif(
        True,
        reason=(
            "Package permissions are getting reworked in "
            "https://github.com/saltstack/salt/pull/66218"
        ),
    ),
]


@pytest.fixture
def salt_systemd_setup(
    salt_call_cli,
    install_salt,
):
    """
    Fixture to set systemd for salt packages to enabled and active
    Note: assumes Salt packages already installed
    """
    install_salt.install()

    # ensure known state, enabled and active
    test_list = ["salt-api", "salt-minion", "salt-master"]
    for test_item in test_list:
        test_cmd = f"systemctl enable {test_item}"
        ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
        assert ret.returncode == 0

        test_cmd = f"systemctl restart {test_item}"
        ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
        assert ret.returncode == 0


def test_salt_systemd_disabled_preservation(
    salt_call_cli, install_salt, salt_systemd_setup
):
    """
    Test upgrade of Salt packages preserve disabled state of systemd
    """
    if not install_salt.upgrade:
        pytest.skip("Not testing an upgrade, do not run")

    # setup systemd to enabled and active for Salt packages
    # pylint: disable=pointless-statement
    salt_systemd_setup

    # ensure known state, disabled
    test_list = ["salt-api", "salt-minion", "salt-master"]
    for test_item in test_list:
        test_cmd = f"systemctl disable {test_item}"
        ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
        assert ret.returncode == 0

    # Upgrade Salt (inc. minion, master, etc.) from previous version and test
    # pylint: disable=pointless-statement
    install_salt.install(upgrade=True)
    time.sleep(60)  # give it some time

    # test for disabled systemd state
    test_list = ["salt-api", "salt-minion", "salt-master"]
    for test_item in test_list:
        test_cmd = f"systemctl show -p UnitFileState {test_item}"
        ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
        test_enabled = ret.stdout.strip().split("=")[1].split('"')[0].strip()
        assert ret.returncode == 0
        assert test_enabled == "disabled"


def test_salt_systemd_enabled_preservation(
    salt_call_cli, install_salt, salt_systemd_setup
):
    """
    Test upgrade of Salt packages preserve enabled state of systemd
    """
    if not install_salt.upgrade:
        pytest.skip("Not testing an upgrade, do not run")

    # setup systemd to enabled and active for Salt packages
    # pylint: disable=pointless-statement
    salt_systemd_setup

    # Upgrade Salt (inc. minion, master, etc.) from previous version and test
    # pylint: disable=pointless-statement
    install_salt.install(upgrade=True)
    time.sleep(60)  # give it some time

    # test for enabled systemd state
    test_list = ["salt-api", "salt-minion", "salt-master"]
    for test_item in test_list:
        test_cmd = f"systemctl show -p UnitFileState {test_item}"
        ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
        test_enabled = ret.stdout.strip().split("=")[1].split('"')[0].strip()
        assert ret.returncode == 0
        assert test_enabled == "enabled"


def test_salt_systemd_inactive_preservation(
    salt_call_cli, install_salt, salt_systemd_setup
):
    """
    Test upgrade of Salt packages preserve inactive state of systemd
    """
    if not install_salt.upgrade:
        pytest.skip("Not testing an upgrade, do not run")

    # setup systemd to enabled and active for Salt packages
    # pylint: disable=pointless-statement
    salt_systemd_setup

    # ensure known state, disabled
    test_list = ["salt-api", "salt-minion", "salt-master"]
    for test_item in test_list:
        test_cmd = f"systemctl stop {test_item}"
        ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
        assert ret.returncode == 0

    # Upgrade Salt (inc. minion, master, etc.) from previous version and test
    # pylint: disable=pointless-statement
    install_salt.install(upgrade=True)
    time.sleep(60)  # give it some time

    # test for inactive systemd state
    test_list = ["salt-api", "salt-minion", "salt-master"]
    for test_item in test_list:
        test_cmd = f"systemctl is-active {test_item}"
        ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
        test_active = ret.stdout.strip().split()[2].strip('"').strip()
        assert ret.returncode == 1
        assert test_active == "inactive"


def test_salt_systemd_active_preservation(
    salt_call_cli, install_salt, salt_systemd_setup
):
    """
    Test upgrade of Salt packages preserve active state of systemd
    """
    if not install_salt.upgrade:
        pytest.skip("Not testing an upgrade, do not run")

    # setup systemd to enabled and active for Salt packages
    # pylint: disable=pointless-statement
    salt_systemd_setup

    # Upgrade Salt (inc. minion, master, etc.) from previous version and test
    # pylint: disable=pointless-statement
    install_salt.install(upgrade=True)
    time.sleep(60)  # give it some time

    # test for active systemd state
    test_list = ["salt-api", "salt-minion", "salt-master"]
    for test_item in test_list:
        test_cmd = f"systemctl is-active {test_item}"
        ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
        test_active = ret.stdout.strip().split()[2].strip('"').strip()
        assert ret.returncode == 0
        assert test_active == "active"


def test_salt_ownership_permission(salt_call_cli, install_salt, salt_systemd_setup):
    """
    Test upgrade of Salt packages preserve existing ownership
    """
    if not install_salt.upgrade:
        pytest.skip("Not testing an upgrade, do not run")

    # setup systemd to enabled and active for Salt packages
    # pylint: disable=pointless-statement
    salt_systemd_setup

    # test ownership for Minion, Master and Api
    test_list = ["salt-api", "salt-minion", "salt-master"]
    for test_item in test_list:
        if "salt-api" == test_item:
            test_cmd = f"ls -dl /run/{test_item}.pid"
            ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
            test_user = ret.stdout.strip().split()[4]
            assert ret.returncode == 0
            assert test_user == "salt"

            test_cmd = f"ls -dl /run/{test_item}.pid"
            ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
            test_group = ret.stdout.strip().split()[5]
            assert ret.returncode == 0
            assert test_group == "salt"
        else:
            test_name = test_item.strip().split("-")[1]
            test_cmd = f"ls -dl /run/salt/{test_name}"
            ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
            test_user = ret.stdout.strip().split()[4]
            assert ret.returncode == 0
            if test_item == "salt-minion":
                assert test_user == "root"
            else:
                assert test_user == "salt"

            ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
            test_group = ret.stdout.strip().split()[5]
            assert ret.returncode == 0
            if test_item == "salt-minion":
                assert test_group == "root"
            else:
                assert test_group == "salt"

    # create master user, and minion user, change conf, restart and test ownership
    test_master_user = "horse"
    test_minion_user = "donkey"
    ret = salt_call_cli.run("--local", "user.list_users")
    user_list = ret.stdout.strip().split(":")[1]

    if test_master_user not in user_list:
        ret = salt_call_cli.run("--local", "user.add", f"{test_master_user}")

    if test_minion_user not in user_list:
        ret = salt_call_cli.run("--local", "user.add", f"{test_minion_user}")

    ret = salt_call_cli.run(
        "--local", "file.comment_line", "/etc/salt/master", "^user:"
    )
    assert ret.returncode == 0

    ret = salt_call_cli.run(
        "--local", "file.comment_line", "/etc/salt/minion", "^user:"
    )
    assert ret.returncode == 0

    test_string = f"\nuser: {test_master_user}\n"
    ret = salt_call_cli.run("--local", "file.append", "/etc/salt/master", test_string)

    test_string = f"\nuser: {test_minion_user}\n"
    ret = salt_call_cli.run("--local", "file.append", "/etc/salt/minion", test_string)

    # restart and check ownership is correct
    test_list = ["salt-api", "salt-minion", "salt-master"]
    for test_item in test_list:
        test_cmd = f"systemctl restart {test_item}"
        ret = salt_call_cli.run("--local", "cmd.run", test_cmd)

    time.sleep(10)  # allow some time for restart

    # test ownership for Minion, Master and Api - horse and donkey
    test_list = ["salt-api", "salt-minion", "salt-master"]
    for test_item in test_list:
        if "salt-api" == test_item:
            test_cmd = f"ls -dl /run/{test_item}.pid"
            ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
            test_user = ret.stdout.strip().split()[4]
            assert ret.returncode == 0
            assert test_user == f"{test_master_user}"

            test_cmd = f"ls -dl /run/{test_item}.pid"
            ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
            test_group = ret.stdout.strip().split()[5]
            assert ret.returncode == 0
            assert test_group == f"{test_master_user}"
        else:
            test_name = test_item.strip().split("-")[1]
            test_cmd = f"ls -dl /run/salt/{test_name}"
            ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
            test_user = ret.stdout.strip().split()[4]
            assert ret.returncode == 0
            if test_item == "salt-minion":
                assert test_user == f"{test_minion_user}"
            else:
                assert test_user == f"{test_master_user}"

            ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
            test_group = ret.stdout.strip().split()[5]
            assert ret.returncode == 0
            if test_item == "salt-minion":
                assert test_group == f"{test_minion_user}"
            else:
                assert test_group == f"{test_master_user}"

    # Upgrade Salt (inc. minion, master, etc.) from previous version and test
    # pylint: disable=pointless-statement
    install_salt.install(upgrade=True)
    time.sleep(60)  # give it some time

    # test ownership for Minion, Master and Api
    test_list = ["salt-api", "salt-minion", "salt-master"]
    for test_item in test_list:
        if "salt-api" == test_item:
            test_cmd = f"ls -dl /run/{test_item}.pid"
            ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
            test_user = ret.stdout.strip().split()[4]
            assert ret.returncode == 0
            assert test_user == f"{test_master_user}"

            test_cmd = f"ls -dl /run/{test_item}.pid"
            ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
            test_group = ret.stdout.strip().split()[5]
            assert ret.returncode == 0
            assert test_group == f"{test_master_user}"
        else:
            test_name = test_item.strip().split("-")[1]
            test_cmd = f"ls -dl /run/salt/{test_name}"
            ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
            test_user = ret.stdout.strip().split()[4]
            assert ret.returncode == 0
            if test_item == "salt-minion":
                assert test_user == f"{test_minion_user}"
            else:
                assert test_user == f"{test_master_user}"

            ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
            test_group = ret.stdout.strip().split()[5]
            assert ret.returncode == 0
            if test_item == "salt-minion":
                assert test_group == f"{test_minion_user}"
            else:
                assert test_group == f"{test_master_user}"

    # restore to defaults to ensure further tests run fine
    ret = salt_call_cli.run(
        "--local", "file.comment_line", "/etc/salt/master", "^user:"
    )
    assert ret.returncode == 0

    ret = salt_call_cli.run(
        "--local", "file.comment_line", "/etc/salt/minion", "^user:"
    )
    assert ret.returncode == 0

    test_string = "\nuser: salt\n"
    ret = salt_call_cli.run("--local", "file.append", "/etc/salt/master", test_string)

    test_string = "\nuser: root\n"
    ret = salt_call_cli.run("--local", "file.append", "/etc/salt/minion", test_string)

    # restart and check ownership is correct
    test_list = ["salt-api", "salt-minion", "salt-master"]
    for test_item in test_list:
        test_cmd = f"systemctl restart {test_item}"
        ret = salt_call_cli.run("--local", "cmd.run", test_cmd)

    time.sleep(10)  # allow some time for restart



import pytest

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
]


@pytest.mark.parametrize("tgts", (("ssh",), ("regular",), ("ssh", "regular")))
def test_publish(salt_ssh_cli, salt_minion, tgts):
    if len(tgts) > 1:
        tgt = "*"
        exp = {"localhost", salt_minion.id}
    else:
        tgt = "localhost" if "ssh" in tgts else salt_minion.id
        exp = {tgt}
    ret = salt_ssh_cli.run(
        "publish.publish",
        tgt,
        "test.ping",
        ssh_minions="ssh" in tgts,
        regular_minions="regular" in tgts,
    )
    assert ret.returncode == 0
    assert ret.data
    assert set(ret.data) == exp
    for id_ in exp:
        assert ret.data[id_] is True


def test_publish_with_arg(salt_ssh_cli, salt_minion):
    ret = salt_ssh_cli.run(
        "publish.publish",
        "*",
        "test.kwarg",
        arg=["cheese=spam"],
        ssh_minions=True,
        regular_minions=True,
    )
    assert ret.returncode == 0
    assert ret.data
    exp = {salt_minion.id, "localhost"}
    assert set(ret.data) == exp
    for id_ in exp:
        assert ret.data[id_]["cheese"] == "spam"


def test_publish_with_yaml_args(salt_ssh_cli, salt_minion):
    args = ["saltines, si", "crackers, nein", "cheese, indeed"]
    test_args = f'["{args[0]}", "{args[1]}", "{args[2]}"]'
    ret = salt_ssh_cli.run(
        "publish.publish",
        "*",
        "test.arg",
        arg=test_args,
        ssh_minions=True,
        regular_minions=True,
    )
    assert ret.returncode == 0
    assert ret.data
    exp = {salt_minion.id, "localhost"}
    assert set(ret.data) == exp
    for id_ in exp:
        assert ret.data[id_]["args"] == args


@pytest.mark.parametrize("tgts", (("ssh",), ("regular",), ("ssh", "regular")))
def test_full_data(salt_ssh_cli, salt_minion, tgts):
    if len(tgts) > 1:
        tgt = "*"
        exp = {"localhost", salt_minion.id}
    else:
        tgt = "localhost" if "ssh" in tgts else salt_minion.id
        exp = {tgt}
    ret = salt_ssh_cli.run(
        "publish.full_data",
        tgt,
        "test.fib",
        arg=20,
        ssh_minions="ssh" in tgts,
        regular_minions="regular" in tgts,
    )
    assert ret.returncode == 0
    assert ret.data
    assert set(ret.data) == exp
    for id_ in exp:
        assert "ret" in ret.data[id_]
        assert ret.data[id_]["ret"][0] == 6765


def test_full_data_kwarg(salt_ssh_cli, salt_minion):
    ret = salt_ssh_cli.run(
        "publish.full_data",
        "*",
        "test.kwarg",
        arg=["cheese=spam"],
        ssh_minions=True,
        regular_minions=True,
    )
    assert ret.returncode == 0
    assert ret.data
    exp = {"localhost", salt_minion.id}
    assert set(ret.data) == exp
    for id_ in exp:
        assert "ret" in ret.data[id_]
        assert ret.data[id_]["ret"]["cheese"] == "spam"



import pathlib
import shutil

import pytest

from salt.exceptions import CommandExecutionError
from salt.utils.versions import Version

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.destructive_test,
    pytest.mark.requires_network,
]

MAX_NPM_VERSION = "5.0.0"


@pytest.fixture(scope="module")
def install_npm(states):
    try:
        ret = states.pkg.installed(name="npm")
        assert ret.result is True
        # Just name the thing we're looking for
        states.npm  # pylint: disable=pointless-statement
    except (CommandExecutionError, AttributeError, AssertionError) as exc:
        pytest.skip(f"Unable to install npm - {exc}")


@pytest.fixture(scope="module")
def apply_gitconfig_workaround(install_npm):
    # See https://github.com/npm/cli/issues/1673 as to why
    gitconfig = pathlib.Path("~/.gitconfig").expanduser().resolve()
    gitconfig_backup = gitconfig.with_suffix(".bak")
    try:
        if gitconfig.exists():
            shutil.move(str(gitconfig), str(gitconfig_backup))
        gitconfig.write_text(
            """[url "git@github.com:"]\n  insteadOf = git://github.com/\n"""
        )
        yield
    finally:
        if not gitconfig_backup.exists():
            gitconfig.unlink()
        else:
            shutil.move(str(gitconfig), str(gitconfig_backup))


@pytest.fixture
def npm(states, modules, apply_gitconfig_workaround):
    try:
        yield states.npm
    finally:
        for pkg in ("pm2", "request", "grunt"):
            modules.npm.uninstall(pkg)


@pytest.mark.skip_if_not_root
@pytest.mark.timeout_unless_on_windows(120)
def test_removed_installed_cycle(npm, modules):
    project_version = "pm2@5.1.0"
    success = modules.npm.uninstall("pm2")
    assert success, "Unable to uninstall pm2 in prep for tests"

    ret = npm.installed(name=project_version)
    assert ret.result is True, "Failed to states.npm.installed {} - {}".format(
        project_version, ret.comment
    )

    ret = npm.removed(name=project_version)
    assert ret.result is True, "Failed to states.npm.removed {} - {}".format(
        project_version, ret.comment
    )


@pytest.fixture
def npm_version(shell, install_npm):
    ret = shell.run("npm", "-v")
    assert ret.returncode == 0
    return ret.stdout.strip()


@pytest.mark.skip(
    reason="This test tries to install from a git repo using ssh, at least now. Skipping.",
)
@pytest.mark.skip_if_not_root
@pytest.mark.skip_if_binaries_missing("git")
def test_npm_install_url_referenced_package(modules, npm, npm_version, states):
    """
    Determine if URL-referenced NPM module can be successfully installed.
    """
    ret = npm.installed(
        name="request/request#v2.88.2",
        registry="https://registry.npmjs.org/",
    )
    assert ret.result is True
    ret = npm.removed(
        name="git://github.com/request/request",
    )
    assert ret.result is True


@pytest.mark.skip_if_not_root
def test_npm_installed_pkgs(npm):
    """
    Basic test to determine if NPM module successfully installs multiple
    packages.
    """
    ret = npm.installed(
        name="unused",
        pkgs=["pm2@5.1.0", "grunt@1.5.3"],
        registry="https://registry.npmjs.org/",
    )
    assert ret.result is True


def test_npm_cache_clean(npm, npm_version):
    """
    Basic test to determine if NPM successfully cleans its cached packages.
    """
    if Version(npm_version) >= Version(MAX_NPM_VERSION):
        pytest.skip("Skip with npm >= 5.0.0 until #41770 is fixed")
    ret = npm.cache_cleaned(name="unused", force=True)
    assert ret.result is True
